__author__ = "Simon Nilsson"

import warnings

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
import os
import warnings
from typing import Any, Dict

import cv2
import matplotlib.pyplot as plt
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.data import plug_holes_shortest_bout
from simba.utils.enums import TagNames
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (get_fn_ext, get_video_meta_data, read_df,
                                    write_df)

plt.interactive(True)
plt.ioff()
warnings.simplefilter(action="ignore", category=FutureWarning)


class ValidateModelOneVideo(ConfigReader, PlottingMixin, TrainModelMixin):
    """
    Create classifier validation video for a single input video. Results are stored in the
    `project_folder/frames/output/validation directory`.

    .. note::
       For improved run-time, see :meth:`simba.sing_run_model_validation_video_mp.ValidateModelOneVideoMultiprocess` for multiprocess class.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str feature_file_path: path to SimBA file (parquet or CSV) containing pose-estimation and feature fields.
    :param str model_path: path to pickled classifier object
    :param float discrimination_threshold: classification threshold.
    :param int shortest_bout: Allowed classified bout length expressed in milliseconds. E.g., `1000` will shift frames classified
        as containing the behavior, but occuring in a bout shorter than `1000`, from `target present to `target absent`.
    :param str create_gantt:
        If SimBA should create gantt charts alongside the validation video. OPTIONS: 'None', 'Gantt chart: final frame only (slightly faster)',
        'Gantt chart: video'.
    :param dict settings: User style settings for video. E.g., {'pose': True, 'animal_names': True, 'styles': None}

    :example:
    >>> test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
    >>>                                 feature_file_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/features_extracted/SF2.csv',
    >>>                                 model_path='/Users/simon/Desktop/envs/troubleshooting/naresh/models/generated_models/Top.sav',
    >>>                                 discrimination_threshold=0.6,
    >>>                                 shortest_bout=50,
    >>>                                 settings={'pose': True, 'animal_names': True, 'styles': None},
    >>>                                 create_gantt='Gantt chart: final frame only (slightly faster)')
    >>> test.run()
    """

    def __init__(
        self,
        config_path: str,
        feature_file_path: str,
        model_path: str,
        discrimination_threshold: float,
        shortest_bout: int,
        create_gantt: str,
        settings: Dict[str, Any],
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        TrainModelMixin.__init__(self)
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        _, self.feature_filename, ext = get_fn_ext(feature_file_path)
        (
            self.discrimination_threshold,
            self.shortest_bout,
            self.create_gantt,
            self.settings,
        ) = (float(discrimination_threshold), shortest_bout, create_gantt, settings)
        if not os.path.exists(self.single_validation_video_save_dir):
            os.makedirs(self.single_validation_video_save_dir)
        _, _, self.fps = self.read_video_info(video_name=self.feature_filename)
        self.clf_name, self.feature_file_path = (
            os.path.basename(model_path).replace(".sav", ""),
            feature_file_path,
        )
        self.video_path = self.find_video_of_file(self.video_dir, self.feature_filename)
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.vid_output_path = os.path.join(
            self.single_validation_video_save_dir,
            f"{self.feature_filename} {self.clf_name}.avi",
        )
        self.clf_data_save_path = os.path.join(
            self.clf_data_validation_dir, self.feature_filename + ".csv"
        )
        self.clf = read_df(file_path=model_path, file_type="pickle")
        self.in_df = read_df(feature_file_path, self.file_type)

    def __run_clf(self):
        self.data_df = self.drop_bp_cords(df=self.in_df)
        self.prob_col_name = f"Probability_{self.clf_name}"
        self.data_df[self.prob_col_name] = self.clf_predict_proba(
            clf=self.clf,
            x_df=self.data_df,
            model_name=self.clf_name,
            data_path=self.feature_file_path,
        )
        self.data_df[self.clf_name] = np.where(
            self.data_df[self.prob_col_name] > self.discrimination_threshold, 1, 0
        )

    def __plug_bouts(self):
        self.data_df = plug_holes_shortest_bout(
            data_df=self.data_df,
            clf_name=self.clf_name,
            fps=self.fps,
            shortest_bout=self.shortest_bout,
        )

    def __save(self):
        write_df(
            df=self.data_df, file_type=self.file_type, save_path=self.clf_data_save_path
        )
        print(f"Predictions created for video {self.feature_filename}...")

    def __create_video(self):
        cap = cv2.VideoCapture(self.video_path)
        fourcc = cv2.VideoWriter_fourcc(*"mp4v")
        if self.create_gantt == "None":
            writer = cv2.VideoWriter(
                self.vid_output_path,
                fourcc,
                self.fps,
                (self.video_meta_data["width"], self.video_meta_data["height"]),
            )
        else:
            self.bouts_df = self.get_bouts_for_gantt(
                data_df=self.data_df, clf_name=self.clf_name, fps=self.fps
            )
            self.gantt_img = self.create_gantt_img(
                self.bouts_df,
                self.clf_name,
                len(self.data_df),
                self.fps,
                "Behavior gantt chart (entire session)",
            )
            self.gantt_img = self.resize_gantt(
                self.gantt_img, self.video_meta_data["height"]
            )
            writer = cv2.VideoWriter(
                self.vid_output_path,
                fourcc,
                self.fps,
                (
                    int(self.video_meta_data["width"] + self.gantt_img.shape[1]),
                    int(self.video_meta_data["height"]),
                ),
            )

        if self.settings["styles"] is None:
            self.settings["styles"] = {}
            space_scaler, radius_scaler, resolution_scaler, font_scaler = (
                60,
                20,
                1500,
                1.5,
            )
            max_dim = max(self.video_meta_data["width"], self.video_meta_data["height"])
            self.settings["styles"]["circle size"] = int(
                radius_scaler / (resolution_scaler / max_dim)
            )
            self.settings["styles"]["font size"] = float(
                font_scaler / (resolution_scaler / max_dim)
            )
            self.settings["styles"]["space_scale"] = int(
                space_scaler / (resolution_scaler / max_dim)
            )

        frm_cnt, clf_frm_cnt = 0, 0
        while cap.isOpened():
            ret, frame = cap.read()
            if frame is None:
                print(
                    "SIMBA WARNING: Some frames appear to be missing in the video vs the data file."
                )
                break
            clf_val = int(self.data_df.loc[frm_cnt, self.clf_name])
            clf_frm_cnt += clf_val
            if self.settings["pose"]:
                for animal_cnt, (animal_name, animal_data) in enumerate(
                    self.animal_bp_dict.items()
                ):
                    for bp_cnt, bp in enumerate(range(len(animal_data["X_bps"]))):
                        x_header, y_header = (
                            animal_data["X_bps"][bp],
                            animal_data["Y_bps"][bp],
                        )
                        animal_cords = tuple(
                            self.in_df.loc[
                                self.in_df.index[frm_cnt], [x_header, y_header]
                            ]
                        )
                        cv2.circle(
                            frame,
                            (int(animal_cords[0]), int(animal_cords[1])),
                            0,
                            self.clr_lst[animal_cnt][bp_cnt],
                            self.settings["styles"]["circle size"],
                        )
            if self.settings["animal_names"]:
                for animal_cnt, (animal_name, animal_data) in enumerate(
                    self.animal_bp_dict.items()
                ):
                    x_header, y_header = (
                        animal_data["X_bps"][0],
                        animal_data["Y_bps"][0],
                    )
                    animal_cords = tuple(
                        self.in_df.loc[self.in_df.index[frm_cnt], [x_header, y_header]]
                    )
                    cv2.putText(
                        frame,
                        animal_name,
                        (int(animal_cords[0]), int(animal_cords[1])),
                        self.font,
                        self.settings["styles"]["font size"],
                        self.clr_lst[animal_cnt][0],
                        1,
                    )
            target_timer = round((1 / self.fps) * clf_frm_cnt, 2)
            cv2.putText(
                frame,
                "Timer",
                (10, int(self.settings["styles"]["space_scale"])),
                self.font,
                self.settings["styles"]["font size"],
                (0, 255, 0),
                2,
            )
            addSpacer = 2
            cv2.putText(
                frame,
                (f"{self.clf_name} {target_timer}s"),
                (10, self.settings["styles"]["space_scale"] * addSpacer),
                self.font,
                self.settings["styles"]["font size"],
                (0, 0, 255),
                2,
            )
            addSpacer += 1
            cv2.putText(
                frame,
                "Ensemble prediction",
                (10, self.settings["styles"]["space_scale"] * addSpacer),
                self.font,
                self.settings["styles"]["font size"],
                (0, 255, 0),
                2,
            )
            addSpacer += 2
            if clf_val == 1:
                cv2.putText(
                    frame,
                    self.clf_name,
                    (10, +self.settings["styles"]["space_scale"] * addSpacer),
                    self.font,
                    self.settings["styles"]["font size"],
                    (2, 166, 249),
                    2,
                )
                addSpacer += 1
            if self.create_gantt == "Gantt chart: final frame only (slightly faster)":
                frame = np.concatenate((frame, self.gantt_img), axis=1)
            elif self.create_gantt == "Gantt chart: video":
                gantt_img = self.create_gantt_img(
                    self.bouts_df,
                    self.clf_name,
                    frm_cnt,
                    self.fps,
                    "Behavior gantt chart",
                )
                gantt_img = self.resize_gantt(
                    gantt_img, self.video_meta_data["video_height"]
                )
                frame = np.concatenate((frame, gantt_img), axis=1)
            elif self.create_gantt != "None":
                frame = cv2.resize(
                    frame,
                    (
                        int(self.video_meta_data["width"] + self.gantt_img.shape[1]),
                        self.video_meta_data["height"],
                    ),
                )
            writer.write(np.uint8(frame))
            print(
                "Frame created: {} / {}".format(
                    str(frm_cnt + 1), str(len(self.data_df))
                )
            )
            frm_cnt += 1

        cap.release()
        writer.release()
        self.timer.stop_timer()
        stdout_success(
            msg=f"Validation video saved at {self.vid_output_path}",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    def run(self):
        self.__run_clf()
        if self.shortest_bout > 1:
            self.__plug_bouts()
        self.__save()
        self.__create_video()


# test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/troubleshooting/dam_nest-c-only_ryan/project_folder/project_config.ini',
#                              feature_file_path='/Users/simon/Desktop/envs/troubleshooting/dam_nest-c-only_ryan/project_folder/csv/features_extracted/LBNF2_Ctrl_P04_4_2021-03-18_19-49-46c.csv',
#                              model_path='/Users/simon/Desktop/envs/troubleshooting/dam_nest-c-only_ryan/models/dam_in_nest.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt=None)
# test.run()


# test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
#                              feature_file_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/features_extracted/SF2.csv',
#                              model_path='/Users/simon/Desktop/envs/troubleshooting/naresh/models/generated_models/Top.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt='Gantt chart: final frame only (slightly faster)')
# test.run()


# test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                              feature_file_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/features_extracted/Together_1.csv',
#                              model_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/models/generated_models/Attack.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt='Gantt chart: final frame only (slightly faster)')
# test.run()


# test.perform_clf()
# test.plug_small_bouts()
# test.save_classification_data()
# test.create_video()

# test = ValidateModelOneVideo(ini_path='/Users/simon/Desktop/troubleshooting/Zebrafish/project_folder/project_config.ini',
#                  feature_file_path=r'/Users/simon/Desktop/troubleshooting/Zebrafish/project_folder/csv/features_extracted/20200730_AB_7dpf_850nm_0002.csv',
#                  model_path='/Users/simon/Desktop/troubleshooting/Zebrafish/models/generated_models/Rheotaxis.sav',
#                  d_threshold=0,
#                  shortest_bout=50,
#                  create_gantt='Gantt chart: final frame only (slightly faster)')
# test.perform_clf()
# test.plug_small_bouts()
# test.save_classification_data()
# test.create_video()


# inifile = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\DLC_two_mice\project_folder\project_config.ini"
# featuresPath = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\DLC_two_mice\project_folder\csv\features_extracted\Together_1.csv"
# modelPath = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\DLC_two_mice\models\generated_models\Attack.sav"
#
# inifile = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\Zebrafish\project_folder\project_config.ini"
# featuresPath = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\Zebrafish\project_folder\csv\features_extracted\20200730_AB_7dpf_850nm_0002.csv"
# modelPath = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\Zebrafish\models\validations\model_files\Rheotaxis_1.sav"
#
# dt = 0.4
# sb = 67
# generategantt = 'Gantt chart: video'
#
# test = ValidateModelOneVideo(ini_path=inifile,feature_file_path=featuresPath,model_path=modelPath,d_threshold=dt,shortest_bout=sb, create_gantt=generategantt)
# test.perform_clf()
# test.plug_small_bouts()
# test.save_classification_data()
# test.create_video()

# cv2.imshow('Window', frame)
# key = cv2.waitKey(3000)
# if key == 27:
#     cv2.destroyAllWindows()



__author__ = "Simon Nilsson"


import itertools
import os
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import shap
from scipy.stats import f_oneway
from sklearn.ensemble import RandomForestClassifier
from sklearn.inspection import permutation_importance
from statsmodels.stats.libqsturng import psturng
from statsmodels.stats.multicomp import pairwise_tukeyhsd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Methods
from simba.utils.printing import SimbaTimer, stdout_success

FEATURE_NAME = "FEATURE NAME"
FEATURE_IMPORTANCE = "IMPORTANCE"
F_STATISTIC = "F-STATISTIC"
MEASURE = "MEASURE"
P_VALUE = "P-VALUE"
CLUSTER = "CLUSTER"
PAIRED = "cluster_paired"
CORRELATION_METHODS = "correlation_methods"
GINI_IMPORTANCE = "gini_importance"
TUKEY = "tukey_posthoc"
METHOD = "method"
TARGET = "TARGET"
PEARSON = "pearson"
KENDALL = "kendall"
SHAP = "shap"
PLOTS = "plots"
CREATE = "create"
SPEARMAN = "spearman"
MEAN = "MEAN"
STDEV = "STANDARD DEVIATION"
PERMUTATION_IMPORTANCE = "permutation_importance"
DESCRIPTIVE_STATISTICS = "descriptive_statistics"
ANOVA_HEADERS = ["FEATURE NAME", "F-STATISTIC", "P-VALUE"]


class ClusterFrequentistCalculator(UnsupervisedMixin, ConfigReader):
    """
    Class for computing frequentist statitics based on cluster assignment labels (for explainability purposes).

    :param str config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param str data_path: path to pickle holding unsupervised results in ``data_map.yaml`` format.
    :param dict settings: dict holding which statistical tests to use

    :example:
    >>> settings = {'scaled': True, 'ANOVA': True, 'tukey_posthoc': True, 'descriptive_statistics': True}
    >>> calculator = ClusterFrequentistCalculator(config_path='unsupervised/project_folder/project_config.ini', data_path='unsupervised/cluster_models/quizzical_rhodes.pickle', settings=settings)
    >>> calculator.run()
    """

    def __init__(self, config_path: str, data_path: str, settings: dict):
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.settings = settings
        check_file_exist_and_readable(file_path=data_path)
        self.data = self.read_pickle(data_path=data_path)
        self.save_path = os.path.join(
            self.logs_path,
            f"cluster_descriptive_statistics_{self.data[Clustering.CLUSTER_MODEL.value][Unsupervised.HASHED_NAME.value]}_{self.datetime}.xlsx",
        )
        self.check_key_exist_in_object(
            object=self.data, key=Clustering.CLUSTER_MODEL.value
        )

    def run(self):
        self.x_data = self.data[Unsupervised.METHODS.value][
            Unsupervised.SCALED_DATA.value
        ]
        self.cluster_data = self.data[Clustering.CLUSTER_MODEL.value][
            Unsupervised.MODEL.value
        ].labels_
        if not self.settings[Unsupervised.SCALED.value]:
            self.feature_data = self.scaler_inverse_transform(
                scaler=self.data[Unsupervised.METHODS.value][Unsupervised.SCALER.value],
                data=self.x_data,
            )
        self.x_y_df = pd.concat(
            [
                self.x_data,
                pd.DataFrame(
                    self.cluster_data, columns=[CLUSTER], index=self.x_data.index
                ),
            ],
            axis=1,
        )
        self.cluster_cnt = self.get_cluster_cnt(
            data=self.cluster_data,
            clusterer_name=self.data[Clustering.CLUSTER_MODEL.value][
                Unsupervised.HASHED_NAME.value
            ],
            minimum_clusters=2,
        )
        with pd.ExcelWriter(self.save_path, mode="w") as writer:
            pd.DataFrame().to_excel(writer, sheet_name=" ", index=True)
        if self.settings[Methods.ANOVA.value]:
            self.__one_way_anovas()
        if self.settings[DESCRIPTIVE_STATISTICS]:
            self.__descriptive_stats()
        if self.settings[TUKEY]:
            self.__tukey_posthoc()
        self.timer.stop_timer()
        stdout_success(
            msg=f"Cluster statistics complete. Data saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __save_results(self, df: pd.DataFrame, name: str):
        with pd.ExcelWriter(self.save_path, mode="a") as writer:
            df.to_excel(writer, sheet_name=name, index=True)

    def __one_way_anovas(self):
        print("Calculating ANOVAs...")
        timer = SimbaTimer(start=True)
        self.anova_results = pd.DataFrame(columns=ANOVA_HEADERS)
        for feature_name in self.data[Unsupervised.METHODS.value][
            Unsupervised.FEATURE_NAMES.value
        ]:
            stats_data = (
                self.x_y_df[[feature_name, "CLUSTER"]]
                .sort_values(by=["CLUSTER"])
                .values
            )
            stats_data = np.split(
                stats_data[:, 0], np.unique(stats_data[:, 1], return_index=True)[1][1:]
            )
            f_val, p_val = f_oneway(*stats_data)
            self.anova_results.loc[len(self.anova_results)] = [
                feature_name,
                f_val,
                p_val,
            ]
        self.anova_results = self.anova_results.sort_values(by=[P_VALUE]).set_index(
            FEATURE_NAME
        )
        self.anova_results[P_VALUE] = self.anova_results[P_VALUE].round(5)
        self.__save_results(df=self.anova_results, name=Methods.ANOVA.value)
        timer.stop_timer()
        stdout_success(
            msg=f"ANOVAs saved in {self.save_path}", elapsed_time=timer.elapsed_time_str
        )

    def __descriptive_stats(self):
        print("Calculating descriptive statistics..")
        timer = SimbaTimer(start=True)
        self.descriptive_results = []
        for feature_name in self.data[Unsupervised.METHODS.value][
            Unsupervised.FEATURE_NAMES.value
        ]:
            agg = (
                self.x_y_df.groupby([CLUSTER])[feature_name]
                .agg(["mean", "std", "sem"])
                .T
            )
            agg[FEATURE_NAME] = feature_name
            agg = (
                agg.reset_index(drop=False)
                .set_index(FEATURE_NAME)
                .rename(columns={"index": MEASURE})
            )
            self.descriptive_results.append(pd.DataFrame(agg))
        self.descriptive_results = pd.concat(self.descriptive_results, axis=0)
        self.__save_results(df=self.descriptive_results, name=DESCRIPTIVE_STATISTICS)
        timer.stop_timer()
        stdout_success(
            msg=f"Descriptive statistics saved in {self.save_path}",
            elapsed_time=timer.elapsed_time_str,
        )

    def __tukey_posthoc(self):
        print("Calculating tukey posthocs...")
        timer = SimbaTimer(start=True)
        self.post_hoc_results = []
        for feature_name in self.data[Unsupervised.METHODS.value][
            Unsupervised.FEATURE_NAMES.value
        ]:
            data = pairwise_tukeyhsd(self.x_y_df[feature_name], self.x_y_df[CLUSTER])
            df = pd.DataFrame(
                data=data._results_table.data[1:], columns=data._results_table.data[0]
            )
            df[P_VALUE] = psturng(
                np.abs(data.meandiffs / data.std_pairs),
                len(data.groupsunique),
                data.df_total,
            )
            df[FEATURE_NAME] = feature_name
            df = df.reset_index(drop=True).set_index(FEATURE_NAME)
            self.post_hoc_results.append(df)
        self.post_hoc_results = pd.concat(self.post_hoc_results, axis=0)
        self.__save_results(df=self.post_hoc_results, name=TUKEY)
        timer.stop_timer()
        stdout_success(
            msg=f"Tukey post-hocs' statistics saved in {self.save_path}",
            elapsed_time=timer.elapsed_time_str,
        )


class EmbeddingCorrelationCalculator(UnsupervisedMixin, ConfigReader):
    def __init__(self, data_path: str, config_path: str, settings: dict):
        """
        Class for correlating dimensionality reduction features with original features (for explainability purposes)

        :param str config_path: path to SimBA configparser.ConfigParser project_config.ini
        :param str data_path: path to pickle holding unsupervised results in ``data_map.yaml`` format.
        :param dict settings: dict holding which statistical tests to use and how to create plots.

        :Example:
        >>> settings = {'correlation_methods': ['pearson', 'kendall', 'spearman'], 'plots': {'create': True, 'correlations': 'pearson', 'palette': 'jet'}}
        >>> calculator = EmbeddingCorrelationCalculator(config_path='unsupervised/project_folder/project_config.ini', data_path='unsupervised/cluster_models/quizzical_rhodes.pickle', settings=settings)
        >>> calculator.run()
        """

        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        check_file_exist_and_readable(file_path=data_path)
        self.settings, self.data_path = settings, data_path
        self.data = self.read_pickle(data_path=self.data_path)
        self.save_path = os.path.join(
            self.logs_path,
            f"embedding_correlations_{self.data[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}_{self.datetime}.csv",
        )

    def run(self):
        print("Calculating embedding correlations...")
        self.x_df = self.data[Unsupervised.METHODS.value][
            Unsupervised.SCALED_DATA.value
        ]
        self.y_df = pd.DataFrame(
            self.data[Unsupervised.DR_MODEL.value][Unsupervised.MODEL.value].embedding_,
            columns=["X", "Y"],
            index=self.x_df.index,
        )
        results = pd.DataFrame()
        for correlation_method in self.settings[CORRELATION_METHODS]:
            results[f"{correlation_method}_Y"] = self.x_df.corrwith(
                self.y_df["Y"], method=correlation_method
            )
            results[f"{correlation_method}_X"] = self.x_df.corrwith(
                self.y_df["X"], method=correlation_method
            )
        results.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Embedding correlations saved in {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )

        if self.settings[PLOTS][CREATE]:
            print("Creating embedding correlation plots...")
            df = pd.concat([self.x_df, self.y_df], axis=1)
            save_dir = os.path.join(self.logs_path, "embedding_correlation_plots")
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            for feature_cnt, feature_name in enumerate(
                self.data[Unsupervised.METHODS.value][Unsupervised.FEATURE_NAMES.value]
            ):
                color_bar = plt.cm.ScalarMappable(cmap=self.settings[PLOTS]["palette"])
                color_bar.set_array([])
                plot = sns.scatterplot(
                    data=df,
                    x="X",
                    y="Y",
                    hue=feature_name,
                    cmap=self.settings[PLOTS]["palette"],
                )
                plot.get_legend().remove()
                plot.figure.colorbar(color_bar, label=feature)
                plt.suptitle(feature_name, x=0.5, y=0.92)
                save_path = os.path.join(save_dir, f"{feature_name}.png")
                plot.figure.savefig(save_path, bbox_inches="tight")
                plot.clear()
                plt.close()
                print(
                    f"Saving image {str(feature_cnt+1)}/{str(len(df.columns))} ({feature_name})"
                )
        stdout_success(msg=f"Embedding correlation calculations complete")


class ClusterXAICalculator(UnsupervisedMixin, ConfigReader):
    def __init__(self, data_path: str, config_path: str, settings: dict):
        """
        Class for building RF models on top of cluster assignments, and calculating explainability metrics on RF models

        :param str config_path: path to SimBA configparser.ConfigParser project_config.ini
        :param str data_path: path to pickle holding unsupervised results in ``data_map.yaml`` format.
        :param dict settings: dict holding which tests to use.

        :Example:
        >>> settings = {'gini_importance': True, 'permutation_importance': True, 'shap': {'method': 'cluster_paired', 'create': True, 'sample': 100}}
        >>> calculator = ClusterXAICalculator(config_path='unsupervised/project_folder/project_config.ini', data_path='unsupervised/cluster_models/quizzical_rhodes.pickle', settings=settings)
        >>> calculator.run()
        """

        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.settings, self.data_path = settings, data_path
        check_file_exist_and_readable(file_path=data_path)
        self.data = self.read_pickle(data_path=self.data_path)
        self.save_path = os.path.join(
            self.logs_path,
            f"cluster_xai_statistics_{self.data[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}_{self.datetime}.xlsx",
        )

    def run(self):
        self.x_df = self.data[Unsupervised.METHODS.value][
            Unsupervised.SCALED_DATA.value
        ]
        self.cluster_data = self.data[Clustering.CLUSTER_MODEL.value][
            Unsupervised.MODEL.value
        ].labels_
        self.x_y_df = pd.concat(
            [
                self.x_df,
                pd.DataFrame(
                    self.cluster_data, columns=[CLUSTER], index=self.x_df.index
                ),
            ],
            axis=1,
        )
        self.cluster_cnt = self.get_cluster_cnt(
            data=self.cluster_data,
            clusterer_name=self.data[Clustering.CLUSTER_MODEL.value][
                Unsupervised.HASHED_NAME.value
            ],
            minimum_clusters=2,
        )
        with pd.ExcelWriter(self.save_path, mode="w") as writer:
            pd.DataFrame().to_excel(writer, sheet_name=" ", index=True)
        self.__train_rf_models()
        print(self.settings[GINI_IMPORTANCE])
        if self.settings[GINI_IMPORTANCE]:
            self.__gini_importance()
        if self.settings[PERMUTATION_IMPORTANCE]:
            self.__permutation_importance()
        if self.settings[SHAP][CREATE]:
            self.__shap_values()
        self.timer.stop_timer()
        stdout_success(
            msg=f"Cluster XAI complete. Data saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __save_results(self, df: pd.DataFrame, name: str):
        with pd.ExcelWriter(self.save_path, mode="a") as writer:
            df.to_excel(writer, sheet_name=name, index=True)

    def __train_rf_models(self):
        print("Training ML model...")
        rf_clf = RandomForestClassifier(
            n_estimators=100,
            max_features="sqrt",
            n_jobs=-1,
            criterion="gini",
            min_samples_leaf=1,
            bootstrap=True,
            verbose=1,
        )
        self.rf_data = {}
        for clf_cnt, cluster_id in enumerate(self.x_y_df[CLUSTER].unique()):
            print(f"Training model {clf_cnt+1}/{self.cluster_cnt}")
            self.rf_data[cluster_id] = {}
            target_df = self.x_y_df[self.x_y_df[CLUSTER] == cluster_id].drop(
                [CLUSTER], axis=1
            )
            non_target_df = self.x_y_df[self.x_y_df[CLUSTER] != cluster_id].drop(
                [CLUSTER], axis=1
            )
            target_df[TARGET] = 1
            non_target_df[TARGET] = 0
            self.rf_data[cluster_id]["X"] = (
                pd.concat([target_df, non_target_df], axis=0)
                .reset_index(drop=True)
                .sample(frac=1)
            )
            self.rf_data[cluster_id]["Y"] = self.rf_data[cluster_id]["X"].pop(TARGET)
            rf_clf.fit(self.rf_data[cluster_id]["X"], self.rf_data[cluster_id]["Y"])
            self.rf_data[cluster_id][Unsupervised.MODEL.value] = deepcopy(rf_clf)

    def __gini_importance(self):
        print("Calculating cluster gini importances'...")
        timer = SimbaTimer(start=True)
        for cluster_id, cluster_data in self.rf_data.items():
            importances = list(
                cluster_data[Unsupervised.MODEL.value].feature_importances_
            )
            gini_data = [
                (feature, round(importance, 6))
                for feature, importance in zip(
                    self.data[Unsupervised.METHODS.value][
                        Unsupervised.FEATURE_NAMES.value
                    ],
                    importances,
                )
            ]
            df = (
                pd.DataFrame(gini_data, columns=[FEATURE_NAME, FEATURE_IMPORTANCE])
                .sort_values(by=[FEATURE_IMPORTANCE], ascending=False)
                .reset_index(drop=True)
            )
            self.__save_results(df=df, name=f"GINI CLUSTER {str(cluster_id)}")
        timer.stop_timer()
        stdout_success(
            msg=f"Cluster features gini importances' complete",
            elapsed_time=timer.elapsed_time_str,
        )

    def __permutation_importance(self):
        print("Calculating permutation importance...")
        timer = SimbaTimer(start=True)
        for cluster_id, cluster_data in self.rf_data.items():
            p_importances = permutation_importance(
                cluster_data[Unsupervised.MODEL.value],
                cluster_data["X"],
                cluster_data["Y"],
                n_repeats=5,
                random_state=0,
            )
            df = pd.DataFrame(
                np.column_stack(
                    [
                        self.data[Unsupervised.METHODS.value][
                            Unsupervised.FEATURE_NAMES.value
                        ],
                        p_importances.importances_mean,
                        p_importances.importances_std,
                    ]
                ),
                columns=[FEATURE_NAME, MEAN, STDEV],
            )
            df = df.sort_values(by=[MEAN], ascending=False).reset_index(drop=True)
            self.__save_results(df=df, name=f"PERMUTATION CLUSTER {str(cluster_id)}")
        timer.stop_timer()
        stdout_success(
            msg=f"Cluster features permutation importances' complete",
            elapsed_time=timer.elapsed_time_str,
        )

    def __shap_values(self):
        if self.settings[SHAP][METHOD] == PAIRED:
            print("Calculating paired-clusters shap values ...")
            timer = SimbaTimer(start=True)
            cluster_combinations = list(
                itertools.combinations(list(self.rf_data.keys()), 2)
            )
            for cluster_one_id, cluster_two_id in cluster_combinations:
                explainer = shap.TreeExplainer(
                    self.rf_data[cluster_one_id]["MODEL"],
                    data=None,
                    model_output="raw",
                    feature_perturbation="tree_path_dependent",
                )
                if self.settings["shap"]["sample"] > (
                    len(self.rf_data[cluster_one_id]["X"])
                    or len(self.rf_data[cluster_two_id]["X"])
                ):
                    self.settings["shap"]["sample"] = min(
                        len(self.rf_data[cluster_one_id]["X"]),
                        len(self.rf_data[cluster_two_id]["X"]),
                    )
                cluster_one_sample = self.rf_data[cluster_one_id]["X"].sample(
                    self.settings["shap"]["sample"], replace=False
                )
                cluster_two_sample = self.rf_data[cluster_two_id]["X"].sample(
                    self.settings["shap"]["sample"], replace=False
                )
                cluster_one_shap = pd.DataFrame(
                    explainer.shap_values(cluster_one_sample, check_additivity=False)[
                        1
                    ],
                    columns=self.rf_data[cluster_one_id]["X"].columns,
                )
                cluster_two_shap = pd.DataFrame(
                    explainer.shap_values(cluster_two_sample, check_additivity=False)[
                        1
                    ],
                    columns=self.rf_data[cluster_two_id]["X"].columns,
                )
                mean_df_cluster_one, stdev_df_cluster_one = pd.DataFrame(
                    cluster_one_shap.mean(), columns=["MEAN"]
                ), pd.DataFrame(cluster_one_shap.std(), columns=["STDEV"])
                mean_df_cluster_two, stdev_df_cluster_two = pd.DataFrame(
                    cluster_two_shap.mean(), columns=["MEAN"]
                ), pd.DataFrame(cluster_two_shap.std(), columns=["STDEV"])
                results_cluster_one = mean_df_cluster_one.join(
                    stdev_df_cluster_one
                ).sort_values(by="MEAN", ascending=False)
                results_cluster_two = mean_df_cluster_two.join(
                    stdev_df_cluster_two
                ).sort_values(by="MEAN", ascending=False)
                self.__save_results(
                    df=results_cluster_one,
                    name=f"SHAP CLUSTER {str(cluster_one_id)} vs. {str(cluster_two_id)}",
                )
                self.__save_results(
                    df=results_cluster_two,
                    name=f"SHAP CLUSTER {str(cluster_two_id)} vs. {str(cluster_one_id)}",
                )
            timer.stop_timer()
            stdout_success(
                msg=f"Paired clusters SHAP values complete",
                elapsed_time=timer.elapsed_time_str,
            )


# settings = {'gini_importance': True, 'permutation_importance': True, 'shap': {'method': 'cluster_paired', 'create': True, 'sample': 100}}
# test = ClusterXAICalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini',
#                             data_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/quizzical_rhodes.pickle',
#                             settings=settings)
# test.run()


# settings = {'correlation_methods': ['pearson', 'kendall', 'spearman'], 'plots': {'create': True, 'correlations': 'pearson', 'palette': 'jet'}}
# test = EmbeddingCorrelationCalculator(data_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/quizzical_rhodes.pickle',
#                                    config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini',
#                                    settings=settings)
# test.run()


# settings = {'scaled': True, 'ANOVA': True, 'tukey_posthoc': True, 'descriptive_statistics': True}
# test = ClusterFrequentistCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini',
#                                    data_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/quizzical_rhodes.pickle',
#                                    settings=settings)
# test.run()


# settings = {'scaled': True, 'anova': True, 'tukey_posthoc': True, 'descriptive_statistics': True}
# test = ClusterStatisticsCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini',
#                                    data_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/nostalgic_albattani.pickle',
#                                    settings=settings)
# test.run()



from __future__ import division

import glob
import math
import os
from itertools import combinations

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from numba import jit, prange
from scipy import stats
from scipy.signal import find_peaks
from scipy.spatial import ConvexHull
from scipy.spatial.qhull import QhullError
from scipy.stats import zscore
from statsmodels.stats.diagnostic import lilliefors

from simba.enums import Paths
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.read_config_unit_tests import (check_if_filepath_list_is_empty,
                                          read_project_path_and_file_type)
from simba.utils.printing import SimbaTimer
from simba.utils.read_write import (get_fn_ext, read_config_file, read_df,
                                    read_video_info, write_df)

TAIL_BP_NAMES = ["objectA", "peduncle_base"]
CENTER_BP_NAMES = ["midpoint"]
MOUTH = ["mouth"]

ANGULAR_DISPERSION_S = [10, 5, 2, 1, 0.5, 0.25]


class FishFeatureExtractor(ConfigReader, FeatureExtractionMixin):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        self.timer = SimbaTimer()
        self.timer.start_timer()
        self.compass_brackets = ["N", "NE", "E", "SE", "S", "SW", "W", "NW", "N"]
        self.compass_brackets_long = [
            "Direction_N",
            "Direction_NE",
            "Direction_E",
            "Direction_SE",
            "Direction_S",
            "Direction_SW",
            "Direction_W",
            "Direction_NW",
        ]
        self.compass_brackets_digits = ["0", "1", "2", "3", "4", "5", "6", "7", "0"]
        self.config = read_config_file(config_path=config_path)
        self.project_path, self.file_type = read_project_path_and_file_type(
            config=self.config
        )
        self.input_file_dir = os.path.join(
            self.project_path, Paths.OUTLIER_CORRECTED.value
        )
        self.save_dir = os.path.join(
            self.project_path, Paths.FEATURES_EXTRACTED_DIR.value
        )
        self.video_info_path = os.path.join(self.project_path, Paths.VIDEO_INFO.value)
        self.video_info_df = pd.read_csv(self.video_info_path)
        bp_names_path = os.path.join(self.project_path, Paths.BP_NAMES.value)
        self.bp_names = list(pd.read_csv(bp_names_path, header=None)[0])
        self.col_headers_shifted = []
        for bp in self.bp_names:
            self.col_headers_shifted.extend(
                (bp + "_x_shifted", bp + "_y_shifted", bp + "_p_shifted")
            )
        self.x_y_cols = []
        self.x_cols_shifted, self.y_cols_shifted = [], []
        for x_name, y_name in zip(self.x_cols, self.y_cols):
            self.x_y_cols.extend((x_name, y_name))
            self.x_cols_shifted.append(x_name + "_shifted")
            self.y_cols_shifted.append(y_name + "_shifted")

        self.roll_windows_values = [
            75,
            50,
            25,
            20,
            15,
            10,
            4,
            2,
        ]
        self.files_found = glob.glob(
            self.input_file_dir + "/*.{}".format(self.file_type)
        )
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg="SIMBA ERROR: No file in {} directory".format(
                self.input_file_dir
            ),
        )
        print(
            "Extracting features from {} {}...".format(
                str(len(self.files_found)), "file(s)"
            )
        )

        for file_path in self.files_found:
            video_timer = SimbaTimer()
            video_timer.start_timer()
            dir_name, file_name, ext = get_fn_ext(file_path)
            self.save_path = os.path.join(self.save_dir, os.path.basename(file_path))
            video_info, self.px_per_mm, self.fps = read_video_info(
                self.video_info_df, file_name
            )
            self.video_width, self.video_height = (
                video_info["Resolution_width"].values,
                video_info["Resolution_height"].values,
            )
            self.angular_dispersion_windows = []
            for i in range(len(ANGULAR_DISPERSION_S)):
                self.angular_dispersion_windows.append(
                    int(self.fps * ANGULAR_DISPERSION_S[i])
                )

            self.csv_df = (
                read_df(file_path, self.file_type).fillna(0).apply(pd.to_numeric)
            )
            try:
                self.csv_df.columns = self.bp_col_names
            except ValueError:
                msg = f"ERROR: Data contains the following fields: {self.csv_df.columns}. \n SimBA wants to use the following field names {self.bp_header_list}"
                print(msg)
                raise ValueError(msg)

            csv_df_shifted = self.csv_df.shift(periods=1)
            csv_df_shifted.columns = self.col_headers_shifted
            self.csv_df_combined = pd.concat(
                [self.csv_df, csv_df_shifted], axis=1, join="inner"
            ).fillna(0)
            self.calc_X_relative_to_Y_movement()
            self.calc_movement()
            self.calc_X_relative_to_Y_movement_rolling_windows()
            self.calc_velocity()
            self.calc_acceleration()
            self.calc_rotation()
            self.calc_N_degree_direction_switches()
            self.bouts_in_same_direction()
            self.calc_45_degree_direction_switches()
            self.hot_end_encode_compass()
            self.calc_directional_switches_in_rolling_windows()
            self.calc_angular_dispersion()
            self.calc_border_distances()
            self.calc_distances_between_body_part()
            self.calc_convex_hulls()
            self.pose_confidence_probabilities()
            self.distribution_tests()
            self.save_file()
            video_timer.stop_timer()
            print(
                f"Features extracted for video {file_name} (elapsed time {video_timer.elapsed_time_str}s)..."
            )

        self.timer.stop_timer()
        print(
            f"Features extracted for all {str(len(self.files_found))} files, data saved in project_folder/csv/features_extracted directory (elapsed time {self.timer.elapsed_time_str}s)"
        )

    def angle2pt_degrees(self, ax, ay, bx, by):
        angle_degrees = math.degrees(math.atan2(ax - bx, by - ay))
        return angle_degrees + 360 if angle_degrees < 0 else angle_degrees

    def angle2pt_radians(self, degrees):
        angle_radians = degrees * math.pi / 180
        return angle_radians

    def angle2pt_sin(self, angle_radians):
        angle_sin = math.sin(angle_radians)
        return angle_sin

    def angle2pt_cos(self, angle_radians):
        angle_cos = math.cos(angle_radians)
        return angle_cos

    @staticmethod
    @jit(nopython=True)
    def count_values_in_range(data: np.array, ranges: np.array):
        results = np.full((data.shape[0], ranges.shape[0]), 0)
        for i in prange(data.shape[0]):
            for j in prange(ranges.shape[0]):
                lower_bound, upper_bound = ranges[j][0], ranges[j][1]
                results[i][j] = data[i][
                    np.logical_and(data[i] >= lower_bound, data[i] <= upper_bound)
                ].shape[0]
        return results

    @staticmethod
    def convex_hull_calculator_mp(arr: np.array, px_per_mm: float) -> float:
        arr = np.unique(arr, axis=0).astype(int)
        if arr.shape[0] < 3:
            return 0
        for i in range(1, arr.shape[0]):
            if (arr[i] != arr[0]).all():
                try:
                    return ConvexHull(arr, qhull_options="En").area / px_per_mm
                except QhullError:
                    return 0
            else:
                pass
        return 0

    @staticmethod
    @jit(nopython=True)
    def euclidian_distance_calc(bp1xVals, bp1yVals, bp2xVals, bp2yVals):
        return np.sqrt((bp1xVals - bp2xVals) ** 2 + (bp1yVals - bp2yVals) ** 2)

    @staticmethod
    @jit(nopython=True)
    def angular_dispersion(cumsum_cos_np, cumsum_sin_np):
        out_array = np.empty((cumsum_cos_np.shape))
        for index in range(cumsum_cos_np.shape[0]):
            X, Y = cumsum_cos_np[index] / (index + 1), cumsum_sin_np[index] / (
                index + 1
            )
            out_array[index] = math.sqrt(X**2 + Y**2)
        return out_array

    def windowed_frequentist_distribution_tests(
        self, data: np.array, feature_name: str, fps: int
    ):
        """
        Helper to compare feature value distributions in 1s sequential time-bins: Kolmogorov-Smirnov and T-tests
        Compares the feature values against a normal distribution: Lillefors, Shapiro.
        Find the number of peaks in *rolling* 1s long feature window.
        """

        (ks_results,) = (np.full((data.shape[0]), -1.0),)
        t_test_results = np.full((data.shape[0]), -1.0)
        lillefors_results = np.full((data.shape[0]), -1.0)
        shapiro_results = np.full((data.shape[0]), -1.0)
        peak_cnt_results = np.full((data.shape[0]), -1.0)

        for i in range(fps, data.shape[0] - fps, fps):
            bin_1_idx, bin_2_idx = [i - fps, i], [i, i + fps]
            bin_1_data, bin_2_data = (
                data[bin_1_idx[0] : bin_1_idx[1]],
                data[bin_2_idx[0] : bin_2_idx[1]],
            )
            ks_results[i : i + fps + 1] = stats.ks_2samp(
                data1=bin_1_data, data2=bin_2_data
            ).statistic
            t_test_results[i : i + fps + 1] = stats.ttest_ind(
                bin_1_data, bin_2_data
            ).statistic

        for i in range(0, data.shape[0] - fps, fps):
            lillefors_results[i : i + fps + 1] = lilliefors(data[i : i + fps])[0]
            shapiro_results[i : i + fps + 1] = stats.shapiro(data[i : i + fps])[0]

        rolling_idx = np.arange(fps)[None, :] + 1 * np.arange(data.shape[0])[:, None]
        for i in range(rolling_idx.shape[0]):
            bin_start_idx, bin_end_idx = rolling_idx[i][0], rolling_idx[i][-1]
            peaks, _ = find_peaks(data[bin_start_idx:bin_end_idx], height=0)
            peak_cnt_results[i] = len(peaks)

        columns = [
            f"{feature_name}_KS",
            f"{feature_name}_TTEST",
            f"{feature_name}_LILLEFORS",
            f"{feature_name}_SHAPIRO",
            f"{feature_name}_PEAK_CNT",
        ]
        return pd.DataFrame(
            np.column_stack(
                (
                    ks_results,
                    t_test_results,
                    lillefors_results,
                    shapiro_results,
                    peak_cnt_results,
                )
            ),
            columns=columns,
        ).round(4)

    @staticmethod
    @jit(nopython=True)
    def consecutive_frames_in_same_compass_direction(direction: np.array):
        results = np.full((direction.shape[0], 1), -1)
        cnt, results[0], last_direction = 0, 0, direction[0]
        for i in prange(1, direction.shape[0]):
            if direction[i] == last_direction:
                cnt += 1
            else:
                cnt = 0
            results[i] = cnt
            last_direction = direction[i]
        return results.flatten()

    @staticmethod
    @jit(nopython=True, fastmath=True)
    def framewise_degree_shift(clockwise_angle: np.array):
        degree_shift = np.full((clockwise_angle.shape[0], 1), np.nan)
        cnt, degree_shift[0], last_angle = 0, 0, clockwise_angle[0]
        for i in prange(1, clockwise_angle.shape[0]):
            degree_shift[i] = math.atan2(
                math.sin(clockwise_angle[i] - last_angle),
                math.cos(clockwise_angle[i] - last_angle),
            )
        return np.absolute(degree_shift.flatten())

    def bouts_in_same_direction(self):
        self.csv_df_combined["Consecutive_ms_in_same_compass_direction"] = (
            self.consecutive_frames_in_same_compass_direction(
                direction=self.csv_df_combined["Compass_digit"].values.astype(int)
            )
            / self.fps
        )
        self.csv_df_combined["Consecutive_ms_in_same_compass_direction_zscore"] = (
            zscore(
                self.csv_df_combined["Consecutive_ms_in_same_compass_direction"].values
            )
        )
        for window in self.roll_windows_values:
            self.csv_df_combined[f"Unique_compass_directions_in_{window}_window"] = (
                self.csv_df_combined["Compass_digit"]
                .astype(int)
                .rolling(window, min_periods=1)
                .apply(lambda x: len(np.unique(x)))
                .astype(int)
            )
        framewise_degree_shift = pd.Series(
            self.framewise_degree_shift(
                clockwise_angle=self.csv_df_combined["Clockwise_angle_degrees"].values
            )
        )
        for window in self.roll_windows_values:
            self.csv_df_combined[f"Degree_shift_{window}_mean"] = (
                framewise_degree_shift.rolling(window, min_periods=1).mean()
            )
            self.csv_df_combined[f"Degree_shift_{window}_median"] = (
                framewise_degree_shift.rolling(window, min_periods=1).median()
            )
            self.csv_df_combined[f"Degree_shift_{window}_sum"] = (
                framewise_degree_shift.rolling(window, min_periods=1).sum()
            )
            self.csv_df_combined[f"Degree_shift_{window}_std"] = (
                framewise_degree_shift.rolling(window, min_periods=1).std()
            )

    def calc_angular_dispersion(self):
        dispersion_array = self.angular_dispersion(
            self.csv_df_combined["Angle_cos_cumsum"].values,
            self.csv_df_combined["Angle_sin_cumsum"].values,
        )
        self.csv_df_combined["Angular_dispersion"] = dispersion_array

        for win in range(len(self.angular_dispersion_windows)):
            col_name = "Angular_dispersion_window_" + str(
                self.angular_dispersion_windows[win]
            )
            self.csv_df_combined[col_name] = (
                self.csv_df_combined["Angular_dispersion"]
                .rolling(self.angular_dispersion_windows[win], min_periods=1)
                .mean()
            )

    def calc_X_relative_to_Y_movement(self):
        temp_df = pd.DataFrame()
        for bp in range(len(self.x_cols)):
            curr_x_col, curr_x_shifted_col, curr_y_col, curr_y_shifted_col = (
                self.x_cols[bp],
                self.x_cols_shifted[bp],
                self.y_cols[bp],
                self.y_cols_shifted[bp],
            )
            temp_df["x"] = (
                self.csv_df_combined[curr_x_col]
                - self.csv_df_combined[curr_x_shifted_col]
            )
            temp_df["y"] = (
                self.csv_df_combined[curr_y_col]
                - self.csv_df_combined[curr_y_shifted_col]
            )
            temp_df["Movement_{}_X_relative_2_Y".format(bp)] = (
                temp_df["x"] - temp_df["y"]
            )
            temp_df.drop(["x", "y"], axis=1, inplace=True)
        self.csv_df_combined["Movement_X_axis_relative_to_Y_axis"] = temp_df.sum(axis=1)

    def calc_movement(self):
        movement_cols = []
        for bp in self.bp_names:
            self.csv_df_combined[f"{bp}_movement"] = (
                self.euclidian_distance_calc(
                    self.csv_df_combined[f"{bp}_x"].values,
                    self.csv_df_combined[f"{bp}_y"].values,
                    self.csv_df_combined[f"{bp}_x_shifted"].values,
                    self.csv_df_combined[f"{bp}_y_shifted"].values,
                )
                / self.px_per_mm
            )
            movement_cols.append(f"{bp}_movement")
        self.csv_df_combined["Summed_movement"] = self.csv_df_combined[
            movement_cols
        ].sum(axis=1)

        for bp in self.bp_names:
            for window in self.roll_windows_values:
                self.csv_df_combined[f"{bp}_movement_{window}_mean"] = (
                    self.csv_df_combined[f"{bp}_movement"]
                    .rolling(window, min_periods=1)
                    .mean()
                )
                self.csv_df_combined[f"{bp}_movement_{window}_sum"] = (
                    self.csv_df_combined[f"{bp}_movement"]
                    .rolling(window, min_periods=1)
                    .sum()
                )

    def calc_X_relative_to_Y_movement_rolling_windows(self):
        for i in self.roll_windows_values:
            currentColName = f"Movement_X_axis_relative_to_Y_axis_mean_{i}"
            self.csv_df_combined[currentColName] = (
                self.csv_df_combined["Movement_X_axis_relative_to_Y_axis"]
                .rolling(i, min_periods=1)
                .mean()
            )
            currentColName = f"Movement_X_axis_relative_to_Y_axis_sum_{i}"
            self.csv_df_combined[currentColName] = (
                self.csv_df_combined["Movement_X_axis_relative_to_Y_axis"]
                .rolling(i, min_periods=1)
                .sum()
            )

    def calc_directional_switches_in_rolling_windows(self):
        for i in self.roll_windows_values:
            currentColName = f"Number_of_direction_switches_{i}"
            self.csv_df_combined[currentColName] = (
                self.csv_df_combined["Direction_switch"].rolling(i, min_periods=1).sum()
            )
            currentColName = f"Directionality_of_switches_switches_{i}"
            self.csv_df_combined[currentColName] = (
                self.csv_df_combined["Switch_direction_value"]
                .rolling(i, min_periods=1)
                .sum()
            )

    def calc_velocity(self):
        self.velocity_fields = []
        for bp in self.bp_names:
            self.csv_df_combined[f"{bp}_velocity"] = (
                self.csv_df_combined[bp + "_movement"]
                .rolling(int(self.fps), min_periods=1)
                .sum()
            )
            self.velocity_fields.append(bp + "_velocity")
        self.csv_df_combined["Bp_velocity_mean"] = self.csv_df_combined[
            self.velocity_fields
        ].mean(axis=1)
        self.csv_df_combined["Bp_velocity_stdev"] = self.csv_df_combined[
            self.velocity_fields
        ].std(axis=1)
        for i in self.roll_windows_values:
            self.csv_df_combined[f"Minimum_avg_bp_velocity_{i}_window"] = (
                self.csv_df_combined["Bp_velocity_mean"].rolling(i, min_periods=1).min()
            )
            self.csv_df_combined[f"Max_avg_bp_velocity_{i}_window"] = (
                self.csv_df_combined["Bp_velocity_mean"].rolling(i, min_periods=1).max()
            )
            self.csv_df_combined[
                f"Absolute_diff_min_max_avg_bp_velocity_{i}_window"
            ] = abs(
                self.csv_df_combined[f"Minimum_avg_bp_velocity_{i}_window"]
                - self.csv_df_combined[f"Max_avg_bp_velocity_{i}_window"]
            )

    def calc_acceleration(self):
        for i in self.roll_windows_values:
            acceleration_fields = []
            for bp in self.bp_names:
                self.csv_df_combined[f"{bp}_velocity_shifted"] = (
                    self.csv_df_combined[f"{bp}_velocity"]
                    .shift(i)
                    .fillna(self.csv_df_combined[f"{bp}_velocity"])
                )
                self.csv_df_combined[f"{bp}_acceleration_{i}_window"] = (
                    self.csv_df_combined[f"{bp}_velocity"]
                    - self.csv_df_combined[f"{bp}_velocity_shifted"]
                )
                self.csv_df_combined = self.csv_df_combined.drop(
                    [f"{bp}_velocity_shifted"], axis=1
                )
                acceleration_fields.append(f"{bp}_acceleration_{i}_window")
            self.csv_df_combined[f"Bp_acceleration_mean_{i}_window"] = (
                self.csv_df_combined[acceleration_fields].mean(axis=1)
            )
            self.csv_df_combined[f"Bp_acceleration_stdev_{i}_window"] = (
                self.csv_df_combined[acceleration_fields].std(axis=1)
            )
        for i in self.roll_windows_values:
            self.csv_df_combined[f"Min_avg_bp_acceleration_{i}_window"] = (
                self.csv_df_combined[f"Bp_acceleration_mean_{i}_window"]
                .rolling(i, min_periods=1)
                .mean()
            )
            self.csv_df_combined[f"Max_avg_bp_acceleration_{i}_window"] = (
                self.csv_df_combined[f"Bp_acceleration_mean_{i}_window"]
                .rolling(i, min_periods=1)
                .mean()
            )
            self.csv_df_combined[
                f"Absolute_diff_min_max_avg_bp_velocity_{i}_window"
            ] = abs(
                self.csv_df_combined[f"Min_avg_bp_acceleration_{i}_window"]
                - self.csv_df_combined[f"Max_avg_bp_acceleration_{i}_window"]
            )

    def calc_N_degree_direction_switches(self):
        degree_lk_180 = {"N": ["S"], "NE": ["SW"], "E": ["W"], "SE": ["NW"]}
        degree_lk_90 = {
            "N": ["W", "E"],
            "NE": ["NW", "SE"],
            "NW": ["SW", "NE"],
            "SW": ["NW", "SE"],
            "SE": ["NE", "SW"],
            "S": ["W", "E"],
            "E": ["N", "S"],
            "W": ["N", "S"],
        }
        dg_df = pd.DataFrame(self.csv_df_combined["Compass_direction"])
        for window in self.roll_windows_values:
            dg_df[f"Compass_direction_{window}"] = dg_df["Compass_direction"].shift(
                window
            )
            dg_df[f"Compass_direction_{window}"].fillna(
                dg_df["Compass_direction"], inplace=True
            )
            dg_df[f"180_degree_switch_{window}"] = 0
            dg_df[f"90_degree_switch_{window}"] = 0
            for k, v in degree_lk_180.items():
                for value in v:
                    dg_df.loc[
                        (dg_df["Compass_direction"] == k)
                        & (dg_df[f"Compass_direction_{window}"] == value),
                        f"180_degree_switch_{window}",
                    ] = 1
                    dg_df.loc[
                        (dg_df[f"Compass_direction_{window}"] == k)
                        & (dg_df["Compass_direction"] == value),
                        f"180_degree_switch_{window}",
                    ] = 1
            for k, v in degree_lk_90.items():
                for value in v:
                    dg_df.loc[
                        (dg_df["Compass_direction"] == k)
                        & (dg_df[f"Compass_direction_{window}"] == value),
                        f"90_degree_switch_{window}",
                    ] = 1
                    dg_df.loc[
                        (dg_df[f"Compass_direction_{window}"] == k)
                        & (dg_df["Compass_direction"] == value),
                        f"90_degree_switch_{window}",
                    ] = 1
            self.csv_df_combined[f"180_degree_switch_{window}"] = dg_df[
                f"180_degree_switch_{window}"
            ]
            self.csv_df_combined[f"90_degree_switch_{window}"] = dg_df[
                f"90_degree_switch_{window}"
            ]

    def calc_rotation(self):
        self.csv_df_combined["Clockwise_angle_degrees"] = self.csv_df_combined.apply(
            lambda x: self.angle2pt_degrees(
                x[CENTER_BP_NAMES[0] + "_x"],
                x[CENTER_BP_NAMES[0] + "_y"],
                x[TAIL_BP_NAMES[0] + "_x"],
                x[TAIL_BP_NAMES[0] + "_y"],
            ),
            axis=1,
        )
        self.csv_df_combined["Angle_radians"] = self.angle2pt_radians(
            self.csv_df_combined["Clockwise_angle_degrees"]
        )
        self.csv_df_combined["Angle_sin"] = self.csv_df_combined.apply(
            lambda x: self.angle2pt_sin(x["Angle_radians"]), axis=1
        )
        self.csv_df_combined["Angle_cos"] = self.csv_df_combined.apply(
            lambda x: self.angle2pt_cos(x["Angle_radians"]), axis=1
        )
        self.csv_df_combined["Angle_sin_cumsum"] = self.csv_df_combined[
            "Angle_sin"
        ].cumsum()
        self.csv_df_combined["Angle_cos_cumsum"] = self.csv_df_combined[
            "Angle_cos"
        ].cumsum()
        compass_lookup = list(
            round(self.csv_df_combined["Clockwise_angle_degrees"] / 45)
        )
        compass_lookup = [int(i) for i in compass_lookup]
        compasFaceList_bracket, compasFaceList_digit = [], []
        for compasDirection in compass_lookup:
            compasFaceList_bracket.append(self.compass_brackets[compasDirection])
            compasFaceList_digit.append(self.compass_brackets_digits[compasDirection])
        self.csv_df_combined["Compass_direction"] = compasFaceList_bracket
        self.csv_df_combined["Compass_digit"] = compasFaceList_digit
        for i in self.roll_windows_values:
            column_name = f"Mean_angle_time_window_{i}"
            self.csv_df_combined[column_name] = (
                self.csv_df_combined["Clockwise_angle_degrees"]
                .rolling(i, min_periods=1)
                .mean()
            )

    def hot_end_encode_compass(self):
        compass_hot_end = pd.get_dummies(
            self.csv_df_combined["Compass_direction"], prefix="Direction"
        )
        compass_hot_end = compass_hot_end.T.reindex(
            self.compass_brackets_long
        ).T.fillna(0)
        self.csv_df_combined = pd.concat(
            [self.csv_df_combined, compass_hot_end], axis=1
        )

    def calc_45_degree_direction_switches(self):
        self.grouped_df = pd.DataFrame()
        v = (
            self.csv_df_combined["Compass_digit"]
            != self.csv_df_combined["Compass_digit"].shift()
        ).cumsum()
        u = self.csv_df_combined.groupby(v)["Compass_digit"].agg(["all", "count"])
        m = u["all"] & u["count"].ge(1)
        self.grouped_df["groups"] = self.csv_df_combined.groupby(v).apply(
            lambda x: (x.index[0], x.index[-1])
        )[m]
        currdirectionList, DirectionSwitchIndexList, currdirectionListValue = [], [], []
        for indexes, row in self.grouped_df.iterrows():
            currdirectionList.append(
                self.csv_df_combined.loc[row["groups"][0]]["Compass_direction"]
            )
            DirectionSwitchIndexList.append(row["groups"][1])
            currdirectionListValue.append(
                self.csv_df_combined.loc[row["groups"][0]]["Compass_digit"]
            )
        self.grouped_df["Direction_switch"] = currdirectionList
        self.grouped_df["Direction_value"] = currdirectionListValue
        self.csv_df_combined.loc[DirectionSwitchIndexList, "Direction_switch"] = 1
        self.csv_df_combined["Compass_digit_shifted"] = self.csv_df_combined[
            "Compass_digit"
        ].shift(-1)
        self.csv_df_combined = self.csv_df_combined.fillna(0)
        self.csv_df_combined["Switch_direction_value"] = self.csv_df_combined.apply(
            lambda x: self.calc_switch_direction(
                x["Compass_digit_shifted"], x["Compass_digit"]
            ),
            axis=1,
        )

    def calc_switch_direction(self, compass_digit_shifted, compass_digit):
        if (compass_digit_shifted == "0") and (compass_digit == "7"):
            return 1
        else:
            return int(compass_digit_shifted) - int(compass_digit)

    def calc_border_distances(self):
        for bp in self.bp_names:
            self.csv_df_combined[f"{bp}_distance_to_left_border"] = (
                self.csv_df_combined[f"{bp}_x"] / self.px_per_mm
            )
            self.csv_df_combined[f"{bp}_distance_to_right_border"] = (
                self.video_width - self.csv_df_combined[f"{bp}_x"]
            ) / self.px_per_mm
            self.csv_df_combined[f"{bp}_distance_to_top_border"] = (
                self.csv_df_combined[f"{bp}_y"] / self.px_per_mm
            )
            self.csv_df_combined[f"{bp}_distance_to_bottom_border"] = (
                self.video_height - self.csv_df_combined[f"{bp}_y"]
            ) / self.px_per_mm

        for side in ["left", "right", "top", "bottom"]:
            side_col_names = [
                c
                for c in self.csv_df_combined.columns
                if f"distance_to_{side}_border" in c
            ]
            self.csv_df_combined[f"Mean_bp_distance_to_{side}_border"] = (
                self.csv_df_combined[side_col_names].mean(axis=1)
            )
            for window in self.roll_windows_values:
                self.csv_df_combined[f"Mean_bp_distance_to_{side}_border_{window}"] = (
                    self.csv_df_combined[f"Mean_bp_distance_to_{side}_border"]
                    .rolling(window, min_periods=1)
                    .mean()
                )
                self.csv_df_combined[f"Std_bp_distance_to_{side}_border_{window}"] = (
                    self.csv_df_combined[f"Mean_bp_distance_to_{side}_border"]
                    .rolling(window, min_periods=1)
                    .std()
                )
                try:
                    # self.csv_df_combined[f'Kurtosis_bp_distance_to_{side}_border_{window}'] = self.csv_df_combined[f'Mean_bp_distance_to_{side}_border'].rolling(window, min_periods=window).kurt()
                    self.csv_df_combined[
                        f"Skew_bp_distance_to_{side}_border_{window}"
                    ] = (
                        self.csv_df_combined[f"Mean_bp_distance_to_{side}_border"]
                        .rolling(window, min_periods=window)
                        .skew()
                    )
                except:
                    # self.csv_df_combined[f'Kurtosis_bp_distance_to_{side}_border_{window}'] = -1
                    self.csv_df_combined[
                        f"Skew_bp_distance_to_{side}_border_{window}"
                    ] = -1

    def calc_distances_between_body_part(self):
        two_point_combs = np.array(list(combinations(self.bp_names, 2)))
        distance_fields = []
        for bps in two_point_combs:
            self.csv_df_combined[f"Distance_{bps[0]}_{bps[1]}"] = (
                self.euclidian_distance_calc(
                    self.csv_df_combined[bps[0] + "_x"].values,
                    self.csv_df_combined[bps[0] + "_y"].values,
                    self.csv_df_combined[bps[1] + "_x"].values,
                    self.csv_df_combined[bps[1] + "_y"].values,
                )
                / self.px_per_mm
            )
            distance_fields.append(f"Distance_{bps[0]}_{bps[1]}")

        for distance_field in distance_fields:
            for window in self.roll_windows_values:
                self.csv_df_combined[f"{distance_field}_mean_{window}"] = (
                    self.csv_df_combined[distance_field]
                    .rolling(window, min_periods=1)
                    .mean()
                )
                self.csv_df_combined[f"{distance_field}_std_{window}"] = (
                    self.csv_df_combined[distance_field]
                    .rolling(window, min_periods=1)
                    .std()
                )
                try:
                    self.csv_df_combined[f"{distance_field}_skew_{window}"] = (
                        self.csv_df_combined[distance_field]
                        .rolling(window, min_periods=1)
                        .skew()
                    )
                    self.csv_df_combined[f"{distance_field}_kurtosis_{window}"] = (
                        self.csv_df_combined[distance_field]
                        .rolling(window, min_periods=1)
                        .kurt()
                    )
                except:
                    self.csv_df_combined[f"{distance_field}_skew_{window}"] = -1
                    self.csv_df_combined[f"{distance_field}_kurtosis_{window}"] = -1

    def calc_convex_hulls(self):
        fish_array = np.reshape(
            self.csv_df[self.x_y_cols].values, (len(self.csv_df / 2), -1, 2)
        )
        self.csv_df_combined["Convex_hull"] = Parallel(
            n_jobs=-1, verbose=0, backend="threading"
        )(
            delayed(self.convex_hull_calculator_mp)(x, self.px_per_mm)
            for x in fish_array
        )
        for window in self.roll_windows_values:
            self.csv_df_combined[f"Convex_hull_mean_{window}_window"] = (
                self.csv_df_combined["Convex_hull"]
                .rolling(window, min_periods=1)
                .mean()
            )
            self.csv_df_combined[f"Convex_hull_std_{window}_window"] = (
                self.csv_df_combined["Convex_hull"].rolling(window, min_periods=1).std()
            )
            self.csv_df_combined[f"Convex_hull_min_{window}_window"] = (
                self.csv_df_combined["Convex_hull"].rolling(window, min_periods=1).min()
            )
            self.csv_df_combined[f"Convex_hull_max_{window}_window"] = (
                self.csv_df_combined["Convex_hull"].rolling(window, min_periods=1).max()
            )
            self.csv_df_combined[
                f"Absolute_diff_min_max_convex_hull_{window}_window"
            ] = abs(
                self.csv_df_combined[f"Convex_hull_min_{window}_window"]
                - self.csv_df_combined[f"Convex_hull_max_{window}_window"]
            )
            try:
                self.csv_df_combined[f"Convex_hull_skew_{window}"] = (
                    self.csv_df_combined["Convex_hull"]
                    .rolling(window, min_periods=1)
                    .skew()
                )
                self.csv_df_combined[f"Convex_hull_kurtosis_{window}"] = (
                    self.csv_df_combined["Convex_hull"]
                    .rolling(window, min_periods=1)
                    .kurt()
                )
            except:
                self.csv_df_combined[f"Convex_hull_skew_{window}"] = -1
                self.csv_df_combined[f"Convex_hull_kurtosis_{window}"] = -1

    def distribution_tests(self):
        distribution_features = [
            "Mean_bp_distance_to_left_border",
            "Mean_bp_distance_to_right_border",
            "Mean_bp_distance_to_top_border",
            "Mean_bp_distance_to_bottom_border",
            "Bp_velocity_mean",
            "Bp_acceleration_mean_25_window",
            "Clockwise_angle_degrees",
            "Convex_hull",
            "Sum_probabilities",
            "Consecutive_ms_in_same_compass_direction",
        ]

        for feature_name in distribution_features:
            results = self.windowed_frequentist_distribution_tests(
                data=self.csv_df_combined[feature_name].values,
                feature_name=feature_name,
                fps=int(self.fps),
            )
            self.csv_df_combined = pd.concat([self.csv_df_combined, results], axis=1)

    def pose_confidence_probabilities(self):
        self.csv_df_combined["Sum_probabilities"] = self.csv_df_combined[
            self.p_cols
        ].sum(axis=1)
        self.csv_df_combined["Sum_probabilities_deviation"] = (
            self.csv_df_combined["Sum_probabilities"].mean()
            - self.csv_df_combined["Sum_probabilities"]
        )
        p_brackets_results = pd.DataFrame(
            self.count_values_in_range(
                data=self.csv_df_combined.filter(self.p_cols).values,
                ranges=np.array(
                    [
                        [0.0, 0.1],
                        [0.000000000, 0.5],
                        [0.000000000, 0.75],
                        [0.000000000, 0.95],
                        [0.000000000, 0.99],
                    ]
                ),
            ),
            columns=[
                "Low_prob_detections_0.1",
                "Low_prob_detections_0.5",
                "Low_prob_detections_0.75",
                "Low_prob_detections_0.95",
                "Low_prob_detections_0.99",
            ],
        )
        self.csv_df_combined = (
            pd.concat([self.csv_df_combined, p_brackets_results], axis=1)
            .reset_index(drop=True)
            .fillna(0)
        )

    def save_file(self):
        self.csv_df_combined = self.csv_df_combined.drop(
            self.col_headers_shifted, axis=1
        )
        self.csv_df_combined = self.csv_df_combined.drop(
            [
                "Compass_digit_shifted",
                "Direction_switch",
                "Switch_direction_value",
                "Compass_digit",
                "Compass_direction",
                "Angle_sin_cumsum",
                "Angle_cos_cumsum",
            ],
            axis=1,
        ).fillna(0)
        write_df(
            self.csv_df_combined.astype(np.float32), self.file_type, self.save_path
        )


# test = FishFeatureExtractor(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini')



__author__ = "Simon Nilsson"


from enum import Enum

import numpy as np


class Unsupervised(Enum):
    ALL_FEATURES_EX_POSE = "ALL FEATURES (EXCLUDING POSE)"
    DATA_SLICE_SELECTION = "data_slice"
    CLF_SLICE_SELECTION = "clf_slice"
    ALL_FEATURES_EXCLUDING_POSE = "ALL FEATURES (EXCLUDING POSE)"
    ALL_FEATURES_INCLUDING_POSE = "ALL FEATURES (INCLUDING POSE)"
    USER_DEFINED_SET = "USER-DEFINED FEATURE SET"
    NAMES = "NAMES"
    START_FRAME = "START_FRAME"
    END_FRAME = "END_FRAME"
    CLASSIFIER = "CLASSIFIER"
    PROBABILITY = "PROBABILITY"
    FRAME = "FRAME"
    VIDEO = "VIDEO"
    FEATURE_PATH = "feature_path"
    BOUT_AGGREGATION_TYPE = "bout_aggregation_type"
    MIN_BOUT_LENGTH = "min_bout_length"
    N_NEIGHBORS = "n_neighbors"
    HASHED_NAME = "HASH"
    DATA = "DATA"
    RAW = "RAW"
    UMAP = "UMAP"
    HDBSCAN = "HDBSCAN"
    TSNE = "TSNE"
    SCALER_TYPE = "SCALER_TYPE"
    CSV = "CSV"
    FORMAT = "format"
    SCALED_DATA = "SCALED_DATA"
    PARAMETERS = "PARAMETERS"
    METHODS = "METHODS"
    DR_MODEL = "DR_MODEL"
    MODEL = "MODEL"
    MIN_DISTANCE = "min_distance"
    EUCLIDEAN = "euclidean"
    FEATURE_NAMES = "FEATURE_NAMES"
    SPREAD = "spread"
    SCALER = "scaler"
    SCALED = "scaled"
    VARIANCE = "variance"
    HYPERPARAMETERS = [N_NEIGHBORS, MIN_DISTANCE, SPREAD, SCALER, VARIANCE]
    FRAME_FEATURES = "FRAME_FEATURES"
    FRAME_POSE = "FRAME_POSE"
    FRAME_TARGETS = "FRAME_TARGETS"
    BOUTS_FEATURES = "BOUTS_FEATURES"
    BOUTS_TARGETS = "BOUTS_TARGETS"
    DATASET_DATA_FIELDS = [
        FRAME_FEATURES,
        FRAME_POSE,
        FRAME_TARGETS,
        BOUTS_FEATURES,
        BOUTS_TARGETS,
    ]
    MIN_MAX = "MIN-MAX"
    STANDARD = "STANDARD"
    QUANTILE = "QUANTILE"
    LOW_VARIANCE_FIELDS = "LOW_VARIANCE_FIELDS"


class Clustering(Enum):
    ALPHA = "alpha"
    MIN_CLUSTER_SIZE = "min_cluster_size"
    MIN_SAMPLES = "min_samples"
    EPSILON = "cluster_selection_epsilon"
    CLUSTER_MODEL = "CLUSTER_MODEL"


class UMLOptions(Enum):
    FEATURE_SLICE_OPTIONS = [
        Unsupervised.ALL_FEATURES_INCLUDING_POSE.value,
        Unsupervised.ALL_FEATURES_EXCLUDING_POSE.value,
        Unsupervised.USER_DEFINED_SET.value,
    ]
    BOUT_AGGREGATION_METHODS = ["MEAN", "MEDIAN"]
    DATA_FORMATS = ["NONE", "SCALED", "RAW"]
    SAVE_FORMATS = ["CSV", "PICKLE"]
    CORRELATION_OPTIONS = ["SPEARMAN", "PEARSONS", "KENDALL"]
    SCALER_OPTIONS = ["MIN-MAX", "STANDARD", "QUANTILE"]
    CATEGORICAL_OPTIONS = ["VIDEO", "CLASSIFIER", "CLUSTER"]
    CONTINUOUS_OPTIONS = ["START_FRAME", "END_FRAME", "PROBABILITY"]
    SPEED_OPTIONS = [round(x, 1) for x in list(np.arange(0.1, 2.1, 0.1))]
    SHAP_CLUSTER_METHODS = ["Paired clusters"]
    DR_ALGO_OPTIONS = ["UMAP", "TSNE"]
    CLUSTERING_ALGO_OPTIONS = ["HDBSCAN"]
    VARIANCE_OPTIONS = list(range(0, 100, 10))
    GRAPH_CNT = list(range(1, 11))
    SCATTER_SIZE = list(range(10, 110, 10))
    SHAP_SAMPLE_OPTIONS = list(range(100, 1100, 100))
    DATA_TYPES = [
        "CLUSTERER HYPER-PARAMETERS",
        "DIMENSIONALITY REDUCTION HYPER-PARAMETERS",
        "SCALER",
        "SCALED DATA",
        "LOW VARIANCE FIELDS",
        "FEATURE_NAMES",
        "FRAME_FEATURES",
        "FRAME_POSE",
        "FRAME_TARGETS",
        "BOUTS_FEATURES",
        "BOUTS_TARGETS",
    ]



__author__ = "Simon Nilsson"

import io
import os
from typing import Dict, List

import cv2
import matplotlib.pyplot as plt
import numpy as np
import PIL

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_that_column_exist
from simba.utils.enums import Formats
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class TresholdPlotCreatorSingleProcess(ConfigReader, PlottingMixin):
    """
    Create line chart visualizations displaying the classification probabilities of a single classifier.

    .. note::
       `Documentation <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`_.
       For improved run-time, use :meth:`simba.plotting.probability_plot_creator_mp.TresholdPlotCreatorMultiprocess`

    .. image:: _static/img/prob_plot.png
       :width: 300
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param str clf_name: Name of the classifier to create visualizations for
    :param bool frame_setting: When True, SimBA creates indidvidual frames in png format
    :param bool video_setting: When True, SimBA creates compressed video in mp4 format
    :param bool last_image: When True, creates image .png representing last frame of the video.
    :param dict style_attr: User-defined style attributes of the visualization (line size, color etc).
    :param List[str] files_found: Files to create threshold plots for.

    Examples
    -----
    >>> style_attr = {'width': 640, 'height': 480, 'font size': 10, 'line width': 6, 'color': 'blue', 'circle size': 20}
    >>> clf_name='Attack'
    >>> files_found=['/_test/project_folder/csv/machine_results/Together_1.csv']

    >>> threshold_plot_creator = TresholdPlotCreatorSingleProcess(config_path='/_test/project_folder/project_config.ini', frame_setting=False, video_setting=True, last_frame=True, clf_name=clf_name, files_found=files_found, style_attr=style_attr)
    >>> threshold_plot_creator.run()
    """

    def __init__(
        self,
        config_path: str,
        clf_name: str,
        frame_setting: bool,
        video_setting: bool,
        last_image: bool,
        style_attr: Dict[str, int],
        files_found: List[str],
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        self.frame_setting, self.video_setting, self.style_attr, self.last_image = (
            frame_setting,
            video_setting,
            style_attr,
            last_image,
        )
        if (
            (not self.frame_setting)
            and (not self.video_setting)
            and (not self.last_image)
        ):
            raise NoSpecifiedOutputError(
                msg="Please choose to either probability videos, probability frames, or both probability frames and videos."
            )

        self.files_found = files_found
        self.orginal_clf_name = clf_name
        self.clf_name = "Probability_" + self.orginal_clf_name
        self.out_width, self.out_height = (
            self.style_attr["width"],
            self.style_attr["height"],
        )
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        if not os.path.exists(self.probability_plot_dir):
            os.makedirs(self.probability_plot_dir)
        print("Processing {} video(s)...".format(str(len(self.files_found))))
        self.timer = SimbaTimer()
        self.timer.start_timer()

    def run(self):
        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            video_info, self.px_per_mm, fps = self.read_video_info(
                video_name=self.video_name
            )
            data_df = read_df(file_path, self.file_type)
            check_that_column_exist(
                df=data_df, column_name=self.clf_name, file_name=self.video_name
            )
            if self.frame_setting:
                self.save_frame_folder_dir = os.path.join(
                    self.probability_plot_dir,
                    self.video_name + "_" + self.orginal_clf_name,
                )
                if not os.path.exists(self.save_frame_folder_dir):
                    os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                self.save_video_path = os.path.join(
                    self.probability_plot_dir,
                    "{}_{}.mp4".format(self.video_name, self.orginal_clf_name),
                )
                self.writer = cv2.VideoWriter(
                    self.save_video_path,
                    self.fourcc,
                    fps,
                    (self.out_width, self.out_height),
                )

            data_df = data_df[self.clf_name]

            if self.last_image:
                self.make_probability_plot(
                    data=data_df,
                    style_attr=self.style_attr,
                    clf_name=self.clf_name,
                    fps=fps,
                    save_path=os.path.join(
                        self.probability_plot_dir,
                        self.video_name
                        + "_{}_{}.png".format(self.orginal_clf_name, "final_image"),
                    ),
                )

            if self.video_setting or self.frame_setting:
                if self.style_attr["y_max"] == "auto":
                    max_y = np.amax(data_df)
                else:
                    max_y = float(self.style_attr["y_max"])

                y_ticks_locs = y_lbls = np.round(np.linspace(0, max_y, 10), 2)
                for i in range(len(data_df)):
                    p_values = list(data_df.loc[0:i])
                    plt.plot(
                        p_values,
                        color=self.style_attr["color"],
                        linewidth=self.style_attr["line width"],
                    )
                    plt.plot(
                        i,
                        p_values[-1],
                        "o",
                        markersize=self.style_attr["circle size"],
                        color=self.style_attr["color"],
                    )
                    plt.ylim([0, max_y])
                    plt.ylabel(
                        "{} {}".format(self.orginal_clf_name, "probability"),
                        fontsize=self.style_attr["font size"],
                    )
                    x_ticks_locs = x_lbls = np.linspace(0, len(p_values), 5)
                    x_lbls = np.round((x_lbls / fps), 1)
                    plt.xlabel("Time (s)", fontsize=self.style_attr["font size"] + 4)
                    plt.grid()
                    plt.xticks(
                        x_ticks_locs,
                        x_lbls,
                        rotation="horizontal",
                        fontsize=self.style_attr["font size"],
                    )
                    plt.yticks(
                        y_ticks_locs, y_lbls, fontsize=self.style_attr["font size"]
                    )
                    plt.suptitle(
                        self.orginal_clf_name,
                        x=0.5,
                        y=0.92,
                        fontsize=self.style_attr["font size"] + 4,
                    )
                    buffer_ = io.BytesIO()
                    plt.savefig(buffer_, format="png")
                    buffer_.seek(0)
                    image = PIL.Image.open(buffer_)
                    ar = np.asarray(image)
                    open_cv_image = cv2.cvtColor(ar, cv2.COLOR_RGB2BGR)
                    open_cv_image = cv2.resize(
                        open_cv_image, (self.out_width, self.out_height)
                    )
                    frame = np.uint8(open_cv_image)
                    buffer_.close()
                    plt.close()
                    if self.frame_setting:
                        frame_save_path = os.path.join(
                            self.save_frame_folder_dir, str(i) + ".png"
                        )
                        cv2.imwrite(frame_save_path, frame)
                    if self.video_setting:
                        self.writer.write(frame)
                    print(
                        "Probability frame: {} / {}. Video: {} ({}/{})".format(
                            str(i + 1),
                            str(len(data_df)),
                            self.video_name,
                            str(file_cnt + 1),
                            len(self.files_found),
                        )
                    )
                if self.video_setting:
                    self.writer.release()
                video_timer.stop_timer()
                print(
                    "Probability plot for video {} saved (elapsed time: {}s)...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )
        self.timer.stop_timer()
        stdout_success(
            msg=f"All probability visualizations created in project_folder/frames/output/probability_plots directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = TresholdPlotCreatorSingleProcess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                         frame_setting=False,
#                                         video_setting=True,
#                                         last_image=True,
#                                         clf_name='Attack',
#                                         files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                         style_attr={'width': 640, 'height': 480, 'font size': 10, 'line width': 6, 'color': 'blue', 'circle size': 20, 'y_max': 'auto'})
# test.create_plots()


#



import itertools
import os
import re
import threading
import time
from configparser import ConfigParser
from copy import deepcopy
from threading import Thread

import cv2
import numpy as np
import pandas as pd

from simba.roi_tools.ROI_move_shape import move_edge, move_edge_align
from simba.roi_tools.ROI_zoom import zoom_in
from simba.utils.data import add_missing_ROI_cols
from simba.utils.enums import ConfigKey, Keys, Paths
from simba.utils.read_write import get_fn_ext


class ROI_image_class:
    def __del__(self):
        self.stop_display_image = True
        self.display_image_thread.join()

    def __init__(
        self,
        config_path,
        colors_dict,
        duplicate_jump_size,
        line_type,
        click_sens,
        text_size,
        text_thickness,
    ):
        config = ConfigParser()
        configFile = str(config_path)
        config.read(configFile)
        self.project_path = config.get(
            ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value
        )
        (
            self.duplicate_jump_size,
            self.line_type,
            self.click_sens,
            self.text_size,
            self.text_thickness,
        ) = (duplicate_jump_size, line_type, click_sens, text_size, text_thickness)
        self.colors = colors_dict
        self.select_color = (128, 128, 128)
        self.zoomed_img_ratio = 1
        self.no_shapes = 0
        self.current_zoom = 100
        self.out_rectangles = []
        self.out_circles = []
        self.out_polygon = []
        self.keyWaitTime = 100
        self.keyWaitChar = -1
        self.cap = None
        self.circle_info = None
        self.window_name = ""
        self.working_frame = None
        self.orig_frame = None
        self.stop_display_image = False
        self.draw_rectangle_roi = False
        self.user_rectangle_roi = []
        self.lock = threading.Lock()
        self.display_image_imshow_callback = self.get_x_y_callback
        self.display_image_thread = Thread(target=self.display_image)
        self.display_image_thread.start()

    def reset(self):
        self.stop_display_image = True

    def update_working_frame(self, cap, start_running, new_video_name):
        with self.lock:
            self.cap = deepcopy(self.cap)
            _, new_frame = cap.read()
            self.working_frame = deepcopy(new_frame)
            self.orig_frame = deepcopy(new_frame)
            if new_video_name:
                self.curr_vid_name = new_video_name
                self.window_name = "Define shape for " + new_video_name
                self.select_color = (128, 128, 128)
                self.zoomed_img_ratio = 1
                self.no_shapes = 0
                self.current_zoom = 100
                self.out_rectangles = []
                self.out_circles = []
                self.out_polygon = []
                self.check_if_ROIs_exist()
                self.keyWaitTime = 100
                self.keyWaitChar = -1
        self.stop_display_image = not start_running
        self.display_image_imshow_callback = self.get_x_y_callback

    def check_if_ROIs_exist(self):
        roi_measurement_path = os.path.join(
            self.project_path, "logs", Paths.ROI_DEFINITIONS.value
        )

        if os.path.isfile(roi_measurement_path):
            rectangles_found = pd.read_hdf(
                roi_measurement_path, key=Keys.ROI_RECTANGLES.value
            )
            circles_found = pd.read_hdf(
                roi_measurement_path, key=Keys.ROI_CIRCLES.value
            )
            polygons_found = pd.read_hdf(
                roi_measurement_path, key=Keys.ROI_POLYGONS.value
            )

            if len(rectangles_found) > 0:
                rectangles_found = rectangles_found[
                    rectangles_found["Video"] == self.curr_vid_name
                ]
                rectangles_found = add_missing_ROI_cols(rectangles_found)
                self.out_rectangles = rectangles_found.to_dict(orient="records")

            if len(circles_found) > 0:
                circles_found = circles_found[
                    circles_found["Video"] == self.curr_vid_name
                ]
                circles_found = add_missing_ROI_cols(circles_found)
                self.out_circles = circles_found.to_dict(orient="records")

            if len(polygons_found) > 0:
                polygons_found = polygons_found[
                    polygons_found["Video"] == self.curr_vid_name
                ]
                polygons_found = add_missing_ROI_cols(polygons_found)
                self.out_polygon = polygons_found.to_dict(orient="records")

            self.insert_all_ROIs_into_image()

    def update_frame_no(self, new_frame_no):
        self.cap.set(1, new_frame_no)
        _, self.working_frame = self.cap.read()
        self.insert_all_ROIs_into_image(change_frame_no=True)

    def draw_rectangle(self, rectangle_info):
        self.draw_rectangle_roi = True
        while len(self.user_rectangle_roi) == 0:
            time.sleep(1)
        ROI = self.user_rectangle_roi
        top_left_x, top_left_y = ROI[0], ROI[1]
        width = abs(ROI[0] - (ROI[2] + ROI[0]))
        height = abs(ROI[2] - (ROI[3] + ROI[2]))
        bottom_right_x, bottom_right_y = top_left_x + width, top_left_y + height

        center_tag_loc = (int(top_left_x + width / 2), int(top_left_y + height / 2))
        br_tag = (int(top_left_x + width), int(top_left_y + height))
        tr_tag = (int(top_left_x + width), int(top_left_y))
        bl_tag = (int(top_left_x), int(top_left_y + height))
        top_tag = (int(top_left_x + width / 2), int(top_left_y))
        right_tag = (int(top_left_x + width), int(top_left_y + height / 2))
        left_tag = (int(top_left_x), int(top_left_y + height / 2))
        bottom_tag = (int(top_left_x + width / 2), int(top_left_y + height))
        self.out_rectangles.append(
            {
                "Video": rectangle_info["Video_name"],
                "Shape_type": "Rectangle",
                "Name": rectangle_info["Name"],
                "Color name": rectangle_info["Shape_color_name"],
                "Color BGR": rectangle_info["Shape_color_BGR"],
                "Thickness": rectangle_info["Shape_thickness"],
                "topLeftX": top_left_x,
                "topLeftY": top_left_y,
                "Bottom_right_X": bottom_right_x,
                "Bottom_right_Y": bottom_right_y,
                "width": width,
                "height": height,
                "Tags": {
                    "Center tag": center_tag_loc,
                    "Top left tag": (top_left_x, top_left_y),
                    "Bottom right tag": br_tag,
                    "Top right tag": tr_tag,
                    "Bottom left tag": bl_tag,
                    "Top tag": top_tag,
                    "Right tag": right_tag,
                    "Left tag": left_tag,
                    "Bottom tag": bottom_tag,
                },
                "Ear_tag_size": int(rectangle_info["Shape_ear_tag_size"]),
            }
        )

        self.insert_all_ROIs_into_image()
        self.user_rectangle_roi = []

    def draw_circle_callback(self, event, x, y, flags, param):
        if event == 1:
            if not self.center_status:
                self.center_X, self.center_Y = int(x), int(y)
                cv2.circle(
                    self.working_frame,
                    (self.center_X, self.center_Y),
                    int(self.circle_info["Shape_ear_tag_size"]),
                    self.circle_info["Shape_color_BGR"],
                    -1,
                )
                self.center_status = True
            else:
                self.border_x, self.border_y = int(x), int(y)
                self.radius = int(
                    (
                        np.sqrt(
                            (self.center_X - self.border_x) ** 2
                            + (self.center_Y - self.border_y) ** 2
                        )
                    )
                )
                border_tag = (int(self.center_X - self.radius), self.center_Y)
                self.out_circles.append(
                    {
                        "Video": self.circle_info["Video_name"],
                        "Shape_type": "Circle",
                        "Name": self.circle_info["Name"],
                        "Color name": self.circle_info["Shape_color_name"],
                        "Color BGR": self.circle_info["Shape_color_BGR"],
                        "Thickness": self.circle_info["Shape_thickness"],
                        "centerX": self.center_X,
                        "centerY": self.center_Y,
                        "radius": self.radius,
                        "Tags": {
                            "Center tag": (self.center_X, self.center_Y),
                            "Border tag": border_tag,
                        },
                        "Ear_tag_size": int(self.circle_info["Shape_ear_tag_size"]),
                    }
                )
                self.insert_all_ROIs_into_image()
                self.not_done = False

    def draw_circle(self, circle_info):
        self.circle_info = circle_info
        self.center_status = False
        self.not_done = True
        self.keyWaitTime = 800
        self.display_image_imshow_callback = self.draw_circle_callback
        while True and not self.stop_display_image:
            if self.keyWaitChar == 27:
                break
            time.sleep(0.1)

    def polygon_x_y_callback(self, event, x, y, flags, param):
        if event == 1:
            if isinstance(self.polygon_pts, np.ndarray):
                print("please press draw again")
                return
            self.click_loc = (int(x), int(y))
            cv2.circle(
                self.working_frame,
                (self.click_loc[0], self.click_loc[1]),
                self.draw_info["Shape_thickness"],
                self.draw_info["Shape_color_BGR"],
                -1,
            )
            self.polygon_pts.append([int(x), int(y)])

    def draw_polygon(self):
        self.polygon_pts = []
        self.display_image_imshow_callback = self.polygon_x_y_callback
        while True and not self.stop_display_image:
            if self.keyWaitChar == 27:
                break
            time.sleep(0.1)
        self.polygon_pts = list(
            k for k, _ in itertools.groupby(self.polygon_pts)
        )  # REMOVES DUPLICATES

        self.polygon_pts = np.array(self.polygon_pts).astype("int32")
        self.poly_center = self.polygon_pts.mean(axis=0)
        self.polygon_pts_dict = {}

        for v, p in enumerate(self.polygon_pts):
            self.polygon_pts_dict["Tag_" + str(v)] = (p[0], p[1])
        self.polygon_pts_dict["Center_tag"] = (
            int(self.poly_center[0]),
            int(self.poly_center[1]),
        )

        self.out_polygon.append(
            {
                "Video": self.draw_info["Video_name"],
                "Shape_type": "Polygon",
                "Name": self.draw_info["Name"],
                "Color name": self.draw_info["Shape_color_name"],
                "Color BGR": self.draw_info["Shape_color_BGR"],
                "Thickness": self.draw_info["Shape_thickness"],
                "Center_X": int(self.poly_center[0]),
                "Center_Y": int(self.poly_center[1]),
                "vertices": self.polygon_pts,
                "Tags": self.polygon_pts_dict,
                "Ear_tag_size": int(self.draw_info["Shape_ear_tag_size"]),
            }
        )

        self.insert_all_ROIs_into_image()

    def initiate_draw(self, draw_dict):
        self.keyWaitChar = -1
        self.draw_info = draw_dict
        if self.draw_info["Shape_type"] is "rectangle":
            self.draw_rectangle(self.draw_info)
        if self.draw_info["Shape_type"] is "circle":
            self.draw_circle(self.draw_info)
        if self.draw_info["Shape_type"] is "polygon":
            self.draw_polygon()

        self.all_shape_names = []
        for r in self.out_rectangles:
            self.all_shape_names.append("Rectangle: " + str(r["Name"]))
        for c in self.out_circles:
            self.all_shape_names.append("Circle: " + str(c["Name"]))
        for p in self.out_polygon:
            self.all_shape_names.append("Polygon: " + str(p["Name"]))
        if (len(self.all_shape_names) > 1) and ("None" in self.all_shape_names):
            self.all_shape_names.remove("None")

        return self.all_shape_names

    def get_x_y_callback(self, event, x, y, flags, param):
        if event == 1:
            self.click_loc = (int(x), int(y))
            self.not_done = False

    def interact_functions(self, interact_method, zoom_val=None):
        self.not_done = True

        def recolor_roi_tags():
            self.not_done = True
            if self.closest_roi["Shape_type"] == "Rectangle":
                if self.closest_tag == "Center tag":
                    cv2.rectangle(
                        self.working_frame,
                        (self.closest_roi["topLeftX"], self.closest_roi["topLeftY"]),
                        (
                            self.closest_roi["topLeftX"] + self.closest_roi["width"],
                            self.closest_roi["topLeftY"] + self.closest_roi["height"],
                        ),
                        self.select_color,
                        self.closest_roi["Thickness"],
                    )
                if (
                    (self.closest_tag == "Top tag")
                    or (self.closest_tag == "Top left tag")
                    or (self.closest_tag == "Top right tag")
                ):
                    cv2.line(
                        self.working_frame,
                        (self.closest_roi["topLeftX"], self.closest_roi["topLeftY"]),
                        (
                            self.closest_roi["topLeftX"] + self.closest_roi["width"],
                            self.closest_roi["topLeftY"],
                        ),
                        self.select_color,
                        self.closest_roi["Thickness"],
                    )
                if (
                    (self.closest_tag == "Bottom tag")
                    or (self.closest_tag == "Bottom left tag")
                    or (self.closest_tag == "Bottom right tag")
                ):
                    cv2.line(
                        self.working_frame,
                        self.closest_roi["Tags"]["Bottom left tag"],
                        self.closest_roi["Tags"]["Bottom right tag"],
                        self.select_color,
                        self.closest_roi["Thickness"],
                    )
                if (
                    (self.closest_tag == "Left tag")
                    or (self.closest_tag == "Bottom left tag")
                    or (self.closest_tag == "Top left tag")
                ):
                    cv2.line(
                        self.working_frame,
                        self.closest_roi["Tags"]["Top left tag"],
                        self.closest_roi["Tags"]["Bottom left tag"],
                        self.select_color,
                        self.closest_roi["Thickness"],
                    )
                if (
                    (self.closest_tag == "Right tag")
                    or (self.closest_tag == "Top right tag")
                    or (self.closest_tag == "Bottom right tag")
                ):
                    cv2.line(
                        self.working_frame,
                        self.closest_roi["Tags"]["Top right tag"],
                        self.closest_roi["Tags"]["Bottom right tag"],
                        self.select_color,
                        self.closest_roi["Thickness"],
                    )

            if self.closest_roi["Shape_type"] == "Circle":
                cv2.circle(
                    self.working_frame,
                    (self.closest_roi["centerX"], self.closest_roi["centerY"]),
                    self.closest_roi["radius"],
                    self.select_color,
                    int(self.closest_roi["Thickness"]),
                )

            if self.closest_roi["Shape_type"] == "Polygon":
                if self.closest_tag == "Center_tag":
                    pts = self.closest_roi["vertices"].reshape((-1, 1, 2))
                    cv2.polylines(
                        self.working_frame,
                        [pts],
                        True,
                        self.select_color,
                        int(self.closest_roi["Thickness"]),
                    )
                else:
                    picked_tag_no = int(re.sub("[^0-9]", "", self.closest_tag))
                    border_tag_1, border_tag_2 = "Tag_" + str(
                        picked_tag_no - 1
                    ), "Tag_" + str(picked_tag_no + 1)
                    if border_tag_1 not in self.closest_roi["Tags"]:
                        border_tag_1 = list(self.closest_roi["Tags"].keys())[-1]
                        if border_tag_1 == "Center_tag":
                            border_tag_1 = list(self.closest_roi["Tags"].keys())[-2]
                    if border_tag_2 not in self.closest_roi["Tags"]:
                        border_tag_2 = list(self.closest_roi["Tags"].keys())[0]
                        if border_tag_2 == "Center_tag":
                            border_tag_2 = list(self.closest_roi["Tags"].keys())[1]
                    cv2.line(
                        self.working_frame,
                        self.closest_roi["Tags"][self.closest_tag],
                        self.closest_roi["Tags"][border_tag_1],
                        self.select_color,
                        self.closest_roi["Thickness"],
                    )
                    cv2.line(
                        self.working_frame,
                        self.closest_roi["Tags"][self.closest_tag],
                        self.closest_roi["Tags"][border_tag_2],
                        self.select_color,
                        self.closest_roi["Thickness"],
                    )

            cv2.circle(
                self.working_frame,
                (
                    self.closest_roi["Tags"][self.closest_tag][0],
                    self.closest_roi["Tags"][self.closest_tag][1],
                ),
                self.closest_roi["Ear_tag_size"],
                self.select_color,
                self.closest_roi["Thickness"],
            )

        def find_closest_ROI_tag():
            self.closest_roi, self.closest_tag, self.closest_dist = {}, {}, np.inf
            merged_out = self.out_rectangles + self.out_circles + self.out_polygon
            for s in merged_out:
                for t in s["Tags"]:
                    dist = int(
                        (
                            np.sqrt(
                                (self.click_loc[0] - s["Tags"][t][0]) ** 2
                                + (self.click_loc[1] - s["Tags"][t][1]) ** 2
                            )
                        )
                    )
                    if ((not self.closest_roi) and (dist < self.click_sens)) or (
                        dist < self.closest_dist
                    ):
                        self.closest_roi, self.closest_tag, self.closest_dist = (
                            s,
                            t,
                            dist,
                        )
            if self.closest_dist is not np.inf:
                recolor_roi_tags()

        def check_if_click_is_tag():
            self.click_roi, self.click_tag, self.click_dist = {}, {}, -np.inf
            merged_out = self.out_rectangles + self.out_circles + self.out_polygon
            merged_out[:] = [
                d for d in merged_out if d.get("Name") != self.closest_roi["Name"]
            ]
            for shape in merged_out:
                for tag in shape["Tags"]:
                    distance = int(
                        (
                            np.sqrt(
                                (self.click_loc[0] - shape["Tags"][tag][0]) ** 2
                                + (self.click_loc[1] - shape["Tags"][tag][1]) ** 2
                            )
                        )
                    )
                    if ((not self.click_roi) and (distance < self.click_sens)) or (
                        distance < self.click_dist
                    ):
                        self.click_roi, self.click_tag, self.closest_dist = (
                            shape,
                            tag,
                            distance,
                        )

        if interact_method == "zoom_home":
            self.zoomed_img = self.orig_frame
            self.current_zoom = 100

        if interact_method == "move_shape":
            self.insert_all_ROIs_into_image(ROI_ear_tags=True)
            # initiate_x_y_callback()  # get x y loc ROI tag
            find_closest_ROI_tag()  # find closest ROI tag to x y loc
            # initiate_x_y_callback()  # get x y loc for new x y loc
            check_if_click_is_tag()
            if self.click_roi:
                move_edge_align(
                    self.closest_roi, self.closest_tag, self.click_roi, self.click_tag
                )
            else:
                move_edge(self.closest_roi, self.closest_tag, self.click_loc)
            self.insert_all_ROIs_into_image()  # re-insert ROIs

        if (interact_method == "zoom_in") or (interact_method == "zoom_out"):
            zoom_in(self, zoom_val)

        if interact_method == None:
            self.insert_all_ROIs_into_image(ROI_ear_tags=False)

    def remove_ROI(self, roi_to_delete):
        if roi_to_delete.startswith("Rectangle"):
            rectangle_name = roi_to_delete.split("Rectangle: ")[1]
            self.out_rectangles[:] = [
                d for d in self.out_rectangles if d.get("Name") != rectangle_name
            ]
        if roi_to_delete.startswith("Circle"):
            circle_name = roi_to_delete.split("Circle: ")[1]
            self.out_circles[:] = [
                d for d in self.out_circles if d.get("Name") != circle_name
            ]
        if roi_to_delete.startswith("Polygon"):
            polygon_name = roi_to_delete.split("Polygon: ")[1]
            self.out_polygon[:] = [
                d for d in self.out_polygon if d.get("Name") != polygon_name
            ]
        self.insert_all_ROIs_into_image()

    def insert_all_ROIs_into_image(
        self,
        ROI_ear_tags=False,
        change_frame_no=False,
        show_zoomed_img=False,
        show_size_info=False,
    ):
        self.no_shapes = 0
        if (change_frame_no is False) and (show_zoomed_img is False):
            self.working_frame = deepcopy(self.orig_frame)

        for e in self.out_rectangles:
            self.no_shapes += 1
            if int(e["Thickness"]) == 1:
                cv2.rectangle(
                    self.working_frame,
                    (e["topLeftX"], e["topLeftY"]),
                    (e["topLeftX"] + e["width"], e["topLeftY"] + e["height"]),
                    self.colors[e["Color name"]],
                    int(e["Thickness"]),
                    lineType=4,
                )
            else:
                cv2.rectangle(
                    self.working_frame,
                    (e["topLeftX"], e["topLeftY"]),
                    (e["topLeftX"] + e["width"], e["topLeftY"] + e["height"]),
                    self.colors[e["Color name"]],
                    int(e["Thickness"]),
                    lineType=self.line_type,
                )
            if ROI_ear_tags is True:
                for t in e["Tags"]:
                    cv2.circle(
                        self.working_frame,
                        e["Tags"][t],
                        e["Ear_tag_size"],
                        self.colors[e["Color name"]],
                        -1,
                    )
            if show_size_info is True:
                area_cm = self.rectangle_size_dict["Rectangles"][e["Name"]]["area_cm"]
                width_cm = self.rectangle_size_dict["Rectangles"][e["Name"]]["width_cm"]
                height_cm = self.rectangle_size_dict["Rectangles"][e["Name"]][
                    "height_cm"
                ]
                cv2.putText(
                    self.working_frame,
                    str(height_cm),
                    (
                        int(e["Tags"]["Left tag"][0] + e["Thickness"]),
                        e["Tags"]["Left tag"][1],
                    ),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    self.text_size / 10,
                    self.colors[e["Color name"]],
                    self.text_thickness,
                    lineType=self.line_type,
                )
                cv2.putText(
                    self.working_frame,
                    str(width_cm),
                    (
                        e["Tags"]["Bottom tag"][0],
                        int(e["Tags"]["Bottom tag"][1] - e["Thickness"]),
                    ),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    self.text_size / 10,
                    self.colors[e["Color name"]],
                    self.text_thickness,
                    lineType=self.line_type,
                )
                cv2.putText(
                    self.working_frame,
                    str(area_cm),
                    (e["Tags"]["Center tag"][0], e["Tags"]["Center tag"][1]),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    self.text_size / 10,
                    self.colors[e["Color name"]],
                    self.text_thickness,
                    lineType=self.line_type,
                )

        for c in self.out_circles:
            self.no_shapes += 1
            cv2.circle(
                self.working_frame,
                (c["centerX"], c["centerY"]),
                c["radius"],
                c["Color BGR"],
                int(c["Thickness"]),
                lineType=self.line_type,
            )
            if ROI_ear_tags is True:
                for t in c["Tags"]:
                    cv2.circle(
                        self.working_frame,
                        c["Tags"][t],
                        c["Ear_tag_size"],
                        self.colors[c["Color name"]],
                        -1,
                    )
            if show_size_info is True:
                area_cm = self.circle_size_dict["Circles"][c["Name"]]["area_cm"]
                cv2.putText(
                    self.working_frame,
                    str(area_cm),
                    (c["Tags"]["Center tag"][0], c["Tags"]["Center tag"][1]),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    self.text_size / 10,
                    self.colors[c["Color name"]],
                    self.text_thickness,
                    lineType=self.line_type,
                )

        for pg in self.out_polygon:
            self.no_shapes += 1
            pts = np.array(pg["vertices"]).reshape((-1, 1, 2))
            cv2.polylines(
                self.working_frame,
                [pts],
                True,
                pg["Color BGR"],
                int(pg["Thickness"]),
                lineType=self.line_type,
            )
            if ROI_ear_tags is True:
                for p in pg["Tags"]:
                    cv2.circle(
                        self.working_frame,
                        pg["Tags"][p],
                        pg["Ear_tag_size"],
                        self.colors[pg["Color name"]],
                        -1,
                    )
            if show_size_info is True:
                area_cm = self.polygon_size_dict["Polygons"][pg["Name"]]["area_cm"]
                cv2.putText(
                    self.working_frame,
                    str(area_cm),
                    (pg["Center_X"], pg["Center_Y"]),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    self.text_size / 10,
                    self.colors[pg["Color name"]],
                    self.text_thickness,
                    lineType=self.line_type,
                )

    def destroy_windows(self):
        cv2.destroyWindow(self.window_name)

    def display_image(self):
        self.stop_display_image = False
        while True:
            if self.stop_display_image:
                cv2.destroyAllWindows()
            else:
                if self.working_frame is None:
                    time.sleep(1)
                    continue
                with self.lock:
                    if self.draw_rectangle_roi:
                        self.user_rectangle_roi = cv2.selectROI(
                            self.window_name, self.working_frame
                        )
                    else:
                        cv2.imshow(self.window_name, self.working_frame)
                cv2.setMouseCallback(
                    self.window_name, self.display_image_imshow_callback
                )
                c = cv2.waitKey(self.keyWaitTime)
                if c != -1:
                    self.keyWaitChar = c



import functools
import multiprocessing
import os
from typing import Iterable, List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd
from shapely.geometry import (LineString, MultiLineString, MultiPolygon, Point,
                              Polygon)

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_instance, check_int,
                                check_iterable_length, check_str)
from simba.utils.enums import Defaults, Formats, TextOptions
from simba.utils.errors import CountError, InvalidInputError
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, find_video_of_file,
                                    get_video_meta_data)
from simba.utils.warnings import FrameRangeWarning

ACCEPTED_TYPES = [Polygon, LineString, MultiPolygon, MultiLineString, Point]
FRAME_COUNT = "frame_count"


def geometry_visualizer(
    data: np.ndarray,
    video_path: Union[str, os.PathLike],
    video_temp_dir: Union[str, os.PathLike],
    video_meta_data: dict,
):
    group = int(data[0][-1])
    colors = list(get_color_dict().values())
    start_frm, end_frm = data[0][-2], data[-1][-2]
    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    video_save_path = os.path.join(video_temp_dir, "{}.mp4".format(str(group)))
    video_writer = cv2.VideoWriter(
        video_save_path,
        fourcc,
        video_meta_data["fps"],
        (video_meta_data["width"], video_meta_data["height"]),
    )
    cap = cv2.VideoCapture(video_path)

    for frm_cnt, frm_id in enumerate(range(start_frm, end_frm - 1)):
        cap.set(1, int(frm_id))
        _, img = cap.read()
        for shape_cnt, shape in enumerate(data[frm_cnt][0:-2]):
            if isinstance(shape, Polygon):
                cv2.polylines(
                    img,
                    [np.array(shape.exterior.coords).astype(np.int)],
                    True,
                    (colors[shape_cnt][::-1]),
                    thickness=TextOptions.LINE_THICKNESS.value,
                )
                interior_coords = [
                    np.array(interior.coords, dtype=np.int32).reshape((-1, 1, 2))
                    for interior in shape.interiors
                ]
                for interior in interior_coords:
                    cv2.polylines(
                        img,
                        [interior],
                        isClosed=True,
                        color=(colors[shape_cnt][::-1]),
                        thickness=TextOptions.LINE_THICKNESS.value,
                    )
            if isinstance(shape, LineString):
                cv2.polylines(
                    img,
                    [np.array(shape.coords, dtype=np.int32)],
                    False,
                    (colors[shape_cnt][::-1]),
                    thickness=TextOptions.LINE_THICKNESS.value,
                )
            if isinstance(shape, MultiPolygon):
                for polygon_cnt, polygon in enumerate(shape.geoms):
                    cv2.polylines(
                        img,
                        [
                            np.array(
                                (polygon.convex_hull.exterior.coords), dtype=np.int32
                            )
                        ],
                        True,
                        (colors[shape_cnt + polygon_cnt + 1][::-1]),
                        thickness=TextOptions.LINE_THICKNESS.value,
                    )
            if isinstance(shape, MultiLineString):
                for line_cnt, line in enumerate(shape.geoms):
                    cv2.polylines(
                        img,
                        [np.array(shape[line_cnt].coords, dtype=np.int32)],
                        False,
                        (colors[shape_cnt][::-1]),
                        thickness=TextOptions.LINE_THICKNESS.value,
                    )
            if isinstance(shape, Point):
                cv2.circle(
                    img,
                    (
                        int(np.array(shape.centroid)[0]),
                        int(np.array(shape.centroid)[1]),
                    ),
                    0,
                    colors[shape_cnt][::-1],
                    TextOptions.LINE_THICKNESS.value,
                )
        video_writer.write(img.astype(np.uint8))
        print(f"Creating frame {frm_id} (core: {group})")
    video_writer.release()
    cap.release()

    return group


class GeometryPlotter(ConfigReader, PlottingMixin):
    """
    A class for creating overlay geometry visualization videos based on provided geometries and video name.

    .. note::
       To quickly create geometries on a white background (useful for troubleshooting unexpected geometries), instead use
       :meth:`simba.mixins.geometry_mixin.GeometryMixin.view_shapes`.

    .. image:: _static/img/GeometryPlotter.gif
       :width: 600
       :align: center

    :param config_path: Union[str, os.PathLike]: Path to SimBA configuration file.
    :param geometries: List[List[Union[Polygon, LineString, MultiPolygon, MultiLineString, Point]]]: List of lists of geometries for each frame.
    :param video_name: str: Name of the input video.
    :param core_cnt: int, optional: Number of CPU cores to use for parallel processing. Defaults to -1 represnting all available cores
    :raises InvalidInputError: If the provided geometries contain invalid data types.
    :raises CountError: If the number of shapes in the geometries does not match the number of frames in the video.
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        geometries: List[List[Union[Polygon, LineString]]],
        video_name: str,
        core_cnt: Optional[int] = -1,
        save_path: Optional[Union[str, os.PathLike]] = None,
    ):
        check_file_exist_and_readable(file_path=config_path)
        check_instance(
            source=self.__class__.__name__, instance=geometries, accepted_types=list
        )
        check_iterable_length(
            source=self.__class__.__name__, val=len(geometries), min=1
        )
        check_str(name="Video name", value=video_name)
        shape_types = set()
        for i in geometries:
            shape_types.update(set([type(x) for x in i]))
        for i in shape_types:
            if i not in ACCEPTED_TYPES:
                raise InvalidInputError(
                    msg=f"geometries contain an invalid datatype {i}. Accepted: {ACCEPTED_TYPES}",
                    source=self.__class__.__name__,
                )
        check_int(
            name="CORE COUNT",
            value=core_cnt,
            min_value=-1,
            max_value=find_core_cnt()[0],
            raise_error=True,
        )
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        self.video_path = find_video_of_file(
            video_dir=self.video_dir, filename=video_name
        )
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        for i in range(len(geometries)):
            if len(geometries[i]) != self.video_meta_data[FRAME_COUNT]:
                FrameRangeWarning(
                    msg=f"Geometry {i+1} contains {len(geometries[i])} shapes but video has {self.video_meta_data[FRAME_COUNT]} frames"
                )
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        self.geometries, self.core_cnt, self.video_name = (
            geometries,
            core_cnt,
            video_name,
        )
        self.geometry_dir = os.path.join(
            self.frames_output_dir, "geometry_visualization"
        )
        if not os.path.isdir(self.geometry_dir):
            os.makedirs(self.geometry_dir)
        self.temp_dir = os.path.join(
            self.frames_output_dir, self.geometry_dir, video_name, "temp"
        )
        if not os.path.isdir(self.temp_dir):
            os.makedirs(self.temp_dir)
        if save_path is None:
            self.save_path = os.path.join(self.geometry_dir, f"{video_name}.mp4")
        else:
            check_if_dir_exists(in_dir=os.path.dirname(save_path))
            self.save_path = save_path

    def run(self):
        video_timer = SimbaTimer(start=True)
        data = pd.DataFrame(np.array(self.geometries).T)
        print(data[0][0], data[0][1])
        data, obs_per_split = self.split_and_group_df(
            df=data,
            splits=self.core_cnt,
            include_row_index=True,
            include_split_order=False,
        )
        for i in range(len(data)):
            new_col = np.full(len(data[i]), fill_value=i).reshape(-1, 1)
            data[i] = np.concatenate((data[i], new_col), axis=1)
        with multiprocessing.Pool(
            self.core_cnt, maxtasksperchild=Defaults.LARGE_MAX_TASK_PER_CHILD.value
        ) as pool:
            constants = functools.partial(
                geometry_visualizer,
                video_path=self.video_path,
                video_temp_dir=self.temp_dir,
                video_meta_data=self.video_meta_data,
            )
            for cnt, result in enumerate(pool.imap(constants, data, chunksize=1)):
                print(f"Section {result}/{len(data)} complete...")
            pool.terminate()
            pool.join()

        print(f"Joining {self.video_name} geometry video...")
        concatenate_videos_in_folder(
            in_folder=self.temp_dir, save_path=self.save_path, remove_splits=True
        )
        video_timer.stop_timer()
        stdout_success(
            msg=f"Geometry video {self.save_path} complete!",
            elapsed_time=video_timer.elapsed_time_str,
            source=self.__class__.__name__,
        )



__author__ = "Simon Nilsson"

import os
from copy import deepcopy

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.enums import TagNames
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class InferenceMulticlassBatch(TrainModelMixin, ConfigReader):

    def __init__(self, config_path: str):

        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )

        if len(self.feature_file_paths) == 0:
            raise NoFilesFoundError(
                "Zero files found in the project_folder/csv/features_extracted directory. Create features before running classifier.",
                source=self.__class__.__name__,
            )
        print(
            f"Analyzing {len(self.feature_file_paths)} file(s) with {self.clf_cnt} classifier(s)..."
        )
        self.timer = SimbaTimer(start=True)
        self.model_dict = self.get_model_info(
            config=self.config, model_cnt=self.clf_cnt
        )

    def _create_p_col_headers(self, model_info: dict):
        self.p_col_headers = []
        for v in model_info["classifier_map"].values():
            self.p_col_headers.append(f"Probability_{v}")

    def run(self):
        for file_cnt, file_path in enumerate(self.feature_file_paths):
            video_timer = SimbaTimer(start=True)
            file_name = get_fn_ext(file_path)[1]
            print(f"Analyzing video {file_name}...")
            in_df = read_df(file_path, self.file_type)
            x_df = self.drop_bp_cords(df=in_df, raise_error=False).astype("float32")
            self.check_df_dataset_integrity(
                df=x_df, logs_path=self.logs_path, file_name=file_name
            )
            out_df = deepcopy(in_df)
            for m, m_hyp in self.model_dict.items():
                if not os.path.isfile(m_hyp["model_path"]):
                    raise NoFilesFoundError(
                        msg=f"{m_hyp['model_path']} is not a VALID model file path",
                        source=self.__class__.__name__,
                    )
                self._create_p_col_headers(model_info=m_hyp)
                clf = self.read_pickle(file_path=m_hyp["model_path"])
                probability_df = pd.DataFrame(
                    self.clf_predict_proba(
                        clf=clf,
                        x_df=x_df,
                        data_path=file_path,
                        model_name=m_hyp["model_name"],
                        multiclass=True,
                    ),
                    columns=self.p_col_headers,
                )
                out_df = pd.concat([out_df, probability_df], axis=1)
            file_save_path = os.path.join(
                self.machine_results_dir, f"{file_name}.{self.file_type}"
            )
            write_df(out_df, self.file_type, file_save_path)
            video_timer.stop_timer()
            print(
                f"Predictions created for {file_name} (elapsed time: {video_timer.elapsed_time_str}) ..."
            )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Multi-class machine predictions complete. {len(self.feature_file_paths)} file(s) saved in project_folder/csv/machine_results directory",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# test = InferenceMulticlassBatch(config_path='/Users/simon/Desktop/envs/troubleshooting/multilabel/project_folder/project_config.ini')
# test.run()


# test.clf.n_classes_
# test.model_dict



__author__ = "Simon Nilsson"

import glob
import itertools
import os
from typing import Dict, Optional

import numpy as np
import pandas as pd
from shapely.geometry import Point, Polygon

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.enums import ConfigKey, Dtypes
from simba.utils.errors import BodypartColumnNotFoundError, NoFilesFoundError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_config_entry, read_df
from simba.utils.warnings import NoDataFoundWarning


class ROIAnalyzer(ConfigReader, FeatureExtractionMixin):
    """

    Analyze movements, entries, exits, and time-spent-in user-defined ROIs. Results are stored in the
    'project_folder/logs' directory of the SimBA project.

    :param str ini_path: Path to SimBA project config file in Configparser format.
    :param Optional[str] data_path: Path to folder holding the data used to caluclate ROI aggregate statistics. If None, then `project_folder/
        csv/outlier_corrected_movement_location`. Deafult: None.
    :param Optional[dict] settings: If dict, the animal body-parts and the probability threshold. If None, then the data is read from the
        project_config.ini. Defalt: None.
    :param Optional[bool] calculate_distances: If True, then calculate movements aggregate statistics (distances and velocities) inside ROIs. Results
                                               are saved in ``project_folder/logs/`` directory. Default: False.

    .. note::
       `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    Examples
    ----------
    >>> settings = {'body_parts': {'Simon': 'Ear_left_1', 'JJ': 'Ear_left_2'}, 'threshold': 0.4}
    >>> roi_analyzer = ROIAnalyzer(ini_path='MyProjectConfig', data_path='outlier_corrected_movement_location', settings=settings, calculate_distances=True)
    >>> roi_analyzer.run()
    >>> roi_analyzer.save()
    """

    def __init__(
        self,
        ini_path: str,
        data_path: Optional[str] = None,
        settings: Optional[dict] = None,
        calculate_distances: Optional[bool] = False,
    ):
        ConfigReader.__init__(self, config_path=ini_path)
        FeatureExtractionMixin.__init__(self)
        self.calculate_distances, self.settings = calculate_distances, settings
        if not os.path.exists(self.detailed_roi_data_dir):
            os.makedirs(self.detailed_roi_data_dir)
        if data_path != None:
            self.input_folder = os.path.join(self.project_path, "csv", data_path)
            self.files_found = glob.glob(self.input_folder + "/*." + self.file_type)
            if len(self.files_found) == 0:
                raise NoFilesFoundError(
                    msg=f"No data files found in {self.input_folder}"
                )

        if not self.settings:
            self.roi_config = dict(self.config.items(ConfigKey.ROI_SETTINGS.value))
            if "animal_1_bp" not in self.roi_config.keys():
                raise BodypartColumnNotFoundError(msg="Please analyze ROI data FIRST.")
            self.settings = {}
            self.settings["threshold"] = read_config_entry(
                self.config,
                ConfigKey.ROI_SETTINGS.value,
                ConfigKey.PROBABILITY_THRESHOLD.value,
                Dtypes.FLOAT.value,
                0.00,
            )
            self.settings["body_parts"] = {}
            self.__check_that_roi_config_data_is_valid()
            for animal_name, bp in self.roi_bp_config.items():
                self.settings["body_parts"][animal_name] = bp

        self.body_part_to_animal_lookup = {}
        for animal_cnt, body_part_name in self.settings["body_parts"].items():
            animal_name = self.find_animal_name_from_body_part_name(
                bp_name=body_part_name, bp_dict=self.animal_bp_dict
            )
            self.body_part_to_animal_lookup[animal_cnt] = animal_name

        self.bp_dict, self.bp_names = {}, []
        for animal_name, bp in self.settings["body_parts"].items():
            self.bp_dict[animal_name] = []
            self.bp_dict[animal_name].extend(
                [f'{bp}_{"x"}', f'{bp}_{"y"}', f'{bp}_{"p"}']
            )
            self.bp_names.extend([f'{bp}_{"x"}', f'{bp}_{"y"}', f'{bp}_{"p"}'])
        self.read_roi_data()

    def __check_that_roi_config_data_is_valid(self):
        all_bps = list(set([x[:-2] for x in self.bp_headers]))
        self.roi_bp_config = {}
        for k, v in self.roi_config.items():
            if "".join([i for i in k if not i.isdigit()]) == "animal__bp":
                id = int("".join(c for c in k if c.isdigit())) - 1
                try:
                    self.roi_bp_config[self.multi_animal_id_list[id]] = v
                except:
                    pass
        for animal, bp in self.roi_bp_config.items():
            if bp not in all_bps:
                raise BodypartColumnNotFoundError(
                    msg=f"Project config setting [{ConfigKey.ROI_SETTINGS.value}][{animal}] is not a valid body-part. Please make sure you have analyzed ROI data."
                )

    def __get_bouts(self, lst=None):
        lst = list(lst)
        return lst[0], lst[-1]

    def run(self):
        """
        Method to analyze ROI statistics.

        Returns
        -------
        Attribute: list
            dist_lst, list of pd.DataFrame holding ROI-dependent movement statistics.
        """
        (
            self.time_dict,
            self.entries_dict,
            self.entries_exit_dict,
            self.movement_dict,
        ) = ({}, {}, {}, {})
        for file_path in self.files_found:
            _, video_name, _ = get_fn_ext(file_path)
            (
                self.time_dict[video_name],
                self.entries_dict[video_name],
                self.entries_exit_dict[video_name],
            ) = ({}, {}, {})
            print("Analysing ROI data for video {}...".format(video_name))
            self.video_recs = self.rectangles_df.loc[
                self.rectangles_df["Video"] == video_name
            ]
            self.video_circs = self.circles_df.loc[
                self.circles_df["Video"] == video_name
            ]
            self.video_polys = self.polygon_df.loc[
                self.polygon_df["Video"] == video_name
            ]
            video_shapes = list(
                itertools.chain(
                    self.video_recs["Name"].unique(),
                    self.video_circs["Name"].unique(),
                    self.video_polys["Name"].unique(),
                )
            )

            if video_shapes == 0:
                NoDataFoundWarning(
                    msg=f"Skipping video {video_name}: No user-defined ROI data found for this video..."
                )
                continue

            else:
                video_settings, pix_per_mm, self.fps = self.read_video_info(
                    video_name=video_name
                )
                self.data_df = read_df(file_path, self.file_type).reset_index(drop=True)
                self.data_df.columns = self.bp_headers
                data_df_sliced = self.data_df[self.bp_names]
                self.video_length_s = data_df_sliced.shape[0] / self.fps
                for animal_name in self.bp_dict:
                    animal_df = self.data_df[self.bp_dict[animal_name]]
                    (
                        self.time_dict[video_name][animal_name],
                        self.entries_dict[video_name][animal_name],
                    ) = ({}, {})
                    self.entries_exit_dict[video_name][animal_name] = {}
                    for _, row in self.video_recs.iterrows():
                        top_left_x, top_left_y, shape_name = (
                            row["topLeftX"],
                            row["topLeftY"],
                            row["Name"],
                        )
                        self.entries_exit_dict[video_name][animal_name][shape_name] = {}
                        bottom_right_x, bottom_right_y = (
                            row["Bottom_right_X"],
                            row["Bottom_right_Y"],
                        )
                        slice_x = animal_df[
                            animal_df[self.bp_dict[animal_name][0]].between(
                                top_left_x, bottom_right_x, inclusive=True
                            )
                        ]
                        slice_y = slice_x[
                            slice_x[self.bp_dict[animal_name][1]].between(
                                top_left_y, bottom_right_y, inclusive=True
                            )
                        ]
                        slice = (
                            slice_y[
                                slice_y[self.bp_dict[animal_name][2]]
                                >= self.settings["threshold"]
                            ]
                            .reset_index()
                            .rename(columns={"index": "frame_no"})
                        )
                        bouts = [
                            self.__get_bouts(g)
                            for _, g in itertools.groupby(
                                list(slice["frame_no"]),
                                key=lambda n, c=itertools.count(): n - next(c),
                            )
                        ]
                        self.time_dict[video_name][animal_name][shape_name] = round(
                            len(slice) / self.fps, 3
                        )
                        self.entries_dict[video_name][animal_name][shape_name] = len(
                            bouts
                        )
                        self.entries_exit_dict[video_name][animal_name][shape_name][
                            "Entry_times"
                        ] = list(map(lambda x: x[0], bouts))
                        self.entries_exit_dict[video_name][animal_name][shape_name][
                            "Exit_times"
                        ] = list(map(lambda x: x[1], bouts))

                    for _, row in self.video_circs.iterrows():
                        center_x, center_y, radius, shape_name = (
                            row["centerX"],
                            row["centerY"],
                            row["radius"],
                            row["Name"],
                        )
                        self.entries_exit_dict[video_name][animal_name][shape_name] = {}
                        animal_df["distance"] = np.sqrt(
                            (animal_df[self.bp_dict[animal_name][0]] - center_x) ** 2
                            + (animal_df[self.bp_dict[animal_name][1]] - center_y) ** 2
                        )
                        slice = (
                            animal_df.loc[
                                (animal_df["distance"] <= radius)
                                & (
                                    animal_df[self.bp_dict[animal_name][2]]
                                    >= self.settings["threshold"]
                                )
                            ]
                            .reset_index()
                            .rename(columns={"index": "frame_no"})
                        )
                        bouts = [
                            self.__get_bouts(g)
                            for _, g in itertools.groupby(
                                list(slice["frame_no"]),
                                key=lambda n, c=itertools.count(): n - next(c),
                            )
                        ]
                        self.time_dict[video_name][animal_name][shape_name] = round(
                            len(slice) / self.fps, 3
                        )
                        self.entries_dict[video_name][animal_name][shape_name] = len(
                            bouts
                        )
                        self.entries_exit_dict[video_name][animal_name][shape_name][
                            "Entry_times"
                        ] = list(map(lambda x: x[0], bouts))
                        self.entries_exit_dict[video_name][animal_name][shape_name][
                            "Exit_times"
                        ] = list(map(lambda x: x[1], bouts))

                    for _, row in self.video_polys.iterrows():
                        polygon_shape, shape_name = (
                            Polygon(
                                list(zip(row["vertices"][:, 0], row["vertices"][:, 1]))
                            ),
                            row["Name"],
                        )
                        self.entries_exit_dict[video_name][animal_name][shape_name] = {}
                        points_arr = animal_df[
                            [self.bp_dict[animal_name][0], self.bp_dict[animal_name][1]]
                        ].to_numpy()
                        contains_func = np.vectorize(
                            lambda p: polygon_shape.contains(Point(p)),
                            signature="(n)->()",
                        )
                        inside_frame_no = [
                            j
                            for sub in np.argwhere(contains_func(points_arr))
                            for j in sub
                        ]
                        slice = (
                            animal_df.loc[
                                (animal_df.index.isin(inside_frame_no))
                                & (
                                    animal_df[self.bp_dict[animal_name][2]]
                                    >= self.settings["threshold"]
                                )
                            ]
                            .reset_index()
                            .rename(columns={"index": "frame_no"})
                        )
                        bouts = [
                            self.__get_bouts(g)
                            for _, g in itertools.groupby(
                                list(slice["frame_no"]),
                                key=lambda n, c=itertools.count(): n - next(c),
                            )
                        ]
                        self.time_dict[video_name][animal_name][shape_name] = round(
                            len(slice) / self.fps, 3
                        )
                        self.entries_dict[video_name][animal_name][shape_name] = len(
                            bouts
                        )
                        self.entries_exit_dict[video_name][animal_name][shape_name][
                            "Entry_times"
                        ] = list(map(lambda x: x[0], bouts))
                        self.entries_exit_dict[video_name][animal_name][shape_name][
                            "Exit_times"
                        ] = list(map(lambda x: x[1], bouts))

                if self.calculate_distances:
                    self.movement_dict[video_name] = {}
                    for animal, shape_dicts in self.entries_exit_dict[
                        video_name
                    ].items():
                        self.movement_dict[video_name][animal] = {}
                        for shape_name, shape_data in shape_dicts.items():
                            d = pd.DataFrame.from_dict(
                                shape_data, orient="index"
                            ).T.values.tolist()
                            for entry in d:
                                df = self.data_df[self.bp_dict[animal][0:2]][
                                    self.data_df.index.isin(
                                        list(range(entry[0], entry[1] + 1))
                                    )
                                ]
                                df = self.create_shifted_df(df=df)
                                df["Movement"] = (
                                    np.sqrt(
                                        (df.iloc[:, 0] - df.iloc[:, 2]) ** 2
                                        + (df.iloc[:, 1] - df.iloc[:, 3]) ** 2
                                    )
                                ) / pix_per_mm
                                self.movement_dict[video_name][animal][shape_name] = (
                                    df["Movement"].sum() / 10
                                )
            self.__transpose_dicts_to_dfs()

    def compute_framewise_distance_to_roi_centroids(self):
        """
        Method to compute frame-wise distances between ROI centroids and animal body-parts.

        Returns
        -------
        Attribute: dict
            roi_centroid_distance
        """

        self.roi_centroid_distance = {}
        for file_path in self.files_found:
            _, video_name, _ = get_fn_ext(file_path)
            self.roi_centroid_distance[video_name] = {}
            video_recs = self.rectangles_df.loc[
                self.rectangles_df["Video"] == video_name
            ]
            video_circs = self.circles_df.loc[self.circles_df["Video"] == video_name]
            video_polys = self.polygon_df.loc[self.polygon_df["Video"] == video_name]
            data_df = read_df(file_path, self.file_type).reset_index(drop=True)
            data_df.columns = self.bp_headers
            for animal_name in self.bp_dict:
                self.roi_centroid_distance[video_name][animal_name] = {}
                animal_df = data_df[self.bp_dict[animal_name]]
                for _, row in video_recs.iterrows():
                    center_cord = (
                        (
                            int(
                                row["Bottom_right_Y"]
                                - ((row["Bottom_right_Y"] - row["topLeftY"]) / 2)
                            )
                        ),
                        (
                            int(
                                row["Bottom_right_X"]
                                - ((row["Bottom_right_X"] - row["topLeftX"]) / 2)
                            )
                        ),
                    )
                    self.roi_centroid_distance[video_name][animal_name][row["Name"]] = (
                        np.sqrt(
                            (animal_df[self.bp_dict[animal_name][0]] - center_cord[0])
                            ** 2
                            + (animal_df[self.bp_dict[animal_name][1]] - center_cord[1])
                            ** 2
                        )
                    )

                for _, row in video_circs.iterrows():
                    center_cord = (row["centerX"], row["centerY"])
                    self.roi_centroid_distance[video_name][animal_name][row["Name"]] = (
                        np.sqrt(
                            (animal_df[self.bp_dict[animal_name][0]] - center_cord[0])
                            ** 2
                            + (animal_df[self.bp_dict[animal_name][1]] - center_cord[1])
                            ** 2
                        )
                    )

                for _, row in video_polys.iterrows():
                    polygon_shape = Polygon(
                        list(zip(row["vertices"][:, 0], row["vertices"][:, 1]))
                    )
                    center_cord = polygon_shape.centroid.coords[0]
                    self.roi_centroid_distance[video_name][animal_name][row["Name"]] = (
                        np.sqrt(
                            (animal_df[self.bp_dict[animal_name][0]] - center_cord[0])
                            ** 2
                            + (animal_df[self.bp_dict[animal_name][1]] - center_cord[1])
                            ** 2
                        )
                    )

    def __transpose_dicts_to_dfs(self):
        self.entries_df = pd.DataFrame(columns=["VIDEO", "ANIMAL", "SHAPE", "ENTRIES"])
        for video_name, video_data in self.entries_dict.items():
            for animal_name, animal_data in video_data.items():
                for shape_name, shape_data in animal_data.items():
                    self.entries_df.loc[len(self.entries_df)] = [
                        video_name,
                        animal_name,
                        shape_name,
                        shape_data,
                    ]
        self.entries_df["ANIMAL"] = self.entries_df["ANIMAL"].map(
            self.body_part_to_animal_lookup
        )

        self.time_df = pd.DataFrame(columns=["VIDEO", "ANIMAL", "SHAPE", "TIME"])
        for video_name, video_data in self.time_dict.items():
            for animal_name, animal_data in video_data.items():
                for shape_name, shape_data in animal_data.items():
                    self.time_df.loc[len(self.time_df)] = [
                        video_name,
                        animal_name,
                        shape_name,
                        shape_data,
                    ]
        self.time_df["ANIMAL"] = self.time_df["ANIMAL"].map(
            self.body_part_to_animal_lookup
        )

        self.detailed_df = pd.DataFrame(
            columns=[
                "VIDEO",
                "ANIMAL",
                "BODY-PART",
                "SHAPE",
                "ENTRY FRAMES",
                "EXIT FRAMES",
            ]
        )
        for video_name, video_data in self.entries_exit_dict.items():
            for animal, animal_data in video_data.items():
                body_part = self.settings["body_parts"][animal]
                for shape_name, shape_data in animal_data.items():
                    df = pd.DataFrame.from_dict(shape_data).rename(
                        columns={
                            "Entry_times": "ENTRY FRAMES",
                            "Exit_times": "EXIT FRAMES",
                        }
                    )
                    df["VIDEO"] = video_name
                    df["ANIMAL"] = animal
                    df["BODY-PART"] = body_part
                    df["SHAPE"] = shape_name
                    self.detailed_df = pd.concat([self.detailed_df, df], axis=0)
        self.detailed_df["ANIMAL"] = self.detailed_df["ANIMAL"].map(
            self.body_part_to_animal_lookup
        )
        self.detailed_df = self.detailed_df[
            ["VIDEO", "ANIMAL", "BODY-PART", "SHAPE", "ENTRY FRAMES", "EXIT FRAMES"]
        ]

        if self.calculate_distances:
            self.movements_df = pd.DataFrame(
                columns=["VIDEO", "ANIMAL", "SHAPE", "MOVEMENT INSIDE SHAPE (CM)"]
            )
            for video_name, video_data in self.movement_dict.items():
                for animal_name, animal_data in video_data.items():
                    for shape_name, shape_data in animal_data.items():
                        self.movements_df.loc[len(self.movements_df)] = [
                            video_name,
                            animal_name,
                            shape_name,
                            shape_data,
                        ]
            self.movements_df["ANIMAL"] = self.movements_df["ANIMAL"].map(
                self.body_part_to_animal_lookup
            )

    def save(self):
        """
        Method to save ROI data to disk. ROI latency and ROI entry data is saved in the "project_folder/logs/" directory.
        If ``calculate_distances`` is True, ROI movement data is saved in the "project_folder/logs/" directory.

        Returns
        -------
        None
        """

        self.entries_df.to_csv(
            os.path.join(self.logs_path, f'{"ROI_entry_data"}_{self.datetime}.csv')
        )
        self.time_df.to_csv(
            os.path.join(self.logs_path, f'{"ROI_time_data"}_{self.datetime}.csv')
        )
        self.detailed_df.to_csv(
            os.path.join(self.logs_path, f'{"Detailed_ROI_data"}_{self.datetime}.csv')
        )
        stdout_success(
            msg='ROI time, ROI entry, and Detailed ROI data, have been saved in the "project_folder/logs/" directory in CSV format.'
        )
        if self.calculate_distances:
            self.movements_df.to_csv(
                os.path.join(
                    self.logs_path, f'{"ROI_movement_data"}_{self.datetime}.csv'
                )
            )
            stdout_success(
                msg='ROI movement data saved in the "project_folder/logs/" directory'
            )

        self.timer.stop_timer()
        stdout_success(
            msg="ROI analysis complete", elapsed_time=self.timer.elapsed_time_str
        )


# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/simba_dev/tests/data/test_projects/zebrafish/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    calculate_distances=True)


# settings = {'body_parts': {'animal_1_bp': 'Ear_left_1', 'animal_2_bp': 'Ear_left_2', 'animal_3_bp': 'Ear_right_1',}, 'threshold': 0.4}
# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    settings=settings,
#                    calculate_distances=True)
# test.run()
# test.save()


# settings = {'body_parts': {'Simon': 'Ear_left_1', 'JJ': 'Ear_left_2'}, 'threshold': 0.4}
# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    settings=settings,
#                    calculate_distances=True)
# test.read_roi_dfs()
# test.analyze_ROIs()
# test.save_data()


# settings = {'body_parts': {'animal_1_bp': 'Ear_left_1', 'animal_2_bp': 'Ear_left_2'}, 'threshold': 0.4}
# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    calculate_distances=True)
# test.run()
# test.analyze_ROIs()
# test.save_data()
